#ifndef DV_DATE_H
#define	DV_DATE_H
// $Id: date.h,v 1.18 2004/12/30 11:42:30 dvermeir Exp $

#include	<sys/time.h>

#include	<stdexcept>
#include	<iostream>
#include	<string>

/*! \file
    A simple Dv::Date class that provides a convenient interface to
    system functions such as locatime_r() etc.
*/

namespace Dv {
namespace Util {
class Duration;

//! A runtime exception thrown by some Date operations.
class DateError: public std::runtime_error {
public:
  //! Name of class, is prepended to each message argument of the constructor.
  static const std::string NAME;
  //! Constructor, prepends NAME to message to obtain runtime_error::what().
  DateError(const std::string& message): std::runtime_error(NAME+": "+message) {}
};

//! A simple Date class.
/*! It provides a convenient interface to system functions such as locatime_r() etc.
    Its only data member is a time_t value. Note that only dates since
    Jan 1, 1970 can be represented.

    The default copy constructor and assignment operations are available.
*/ 
class Date {
public:
  //! Default constructor, initializes to ``now''.
  Date();
  //! Date(time_t t) assumes that t is number of seconds since Jan 1, 1970
  Date(time_t secs);	// Date(0) is NULL date
  //! Initialize Date, note that january = 0.
  Date(unsigned int year, unsigned int month, unsigned int day, 
       unsigned int hrs=0, unsigned int min=0, unsigned int sec=0); 

  //! The string s is parsed as a date.
  /*! This constructor uses the gnu getdate parsing function
      which accepts a number of formats, e.g.

      \verbatim
	1970-09-17           # ISO 8601.
  	70-9-17              # This century assumed by default.
  	70-09-17             # Leading zeros are ignored.
  	9/17/72              # Common U.S. writing.
  	24 September 1972
  	24 Sept 72           # September has a special abbreviation.
  	24 Sep 72            # Three-letter abbreviations always allowed.
  	Sep 24, 1972
  	24-sep-72
  	24sep72

        20:02:0
        20:02
  	8:02pm
  	20:02-0500		# In EST (Eastern U.S. Standard Time)
      \endverbatim
      
      and many others (see the gnu filutils documentation)
  */
  Date(const std::string& s) throw (DateError);

  //! Produce a string representing the date, according to format.
  /*! The format is any string acceptable to strftime.
  	If fmt=0, the default is "\%a \%b \%d \%H:\%M:\%S \%Y"
  	e.g. 
        \code
		"Wed Aug 12 18:38:10 1998"
        \endcode
      The default format is acceptable to the parser.
  */
  std::string		str(const char*fmt=0) const;
  //! Like str() but formats in GMT time zone, e.g. 03:00 MET DST is 02:00 GMT.
  std::string		gmt(const char*fmt=0) const;
  //@{
  /*! Inspection functions that return a part of the Date object.
  */
  //! Return number of years.
  unsigned short	year() const;
  //! Return month of date, january = 0.
  unsigned short 	month() const;
  //! Return day in month (1..31).
  unsigned short	day() const;
  //! Return hour of day, between 00 and 23.
  unsigned short	hours() const;
  //! Return minutes in the hour, between 0 and 59.
  unsigned short	minutes() const;
  //! Return seconds in the minute, between 0 and 59.
  unsigned short	seconds() const;
  //! Return weekday, sunday = 0, monday = 1, ..
  unsigned short	wday() const;
  //! Return day in year, between 0 and 365.
  unsigned short	yday() const;
  //! Return number of week (0..53) in year, week 1 starts with first monday,
  unsigned short	yweek() const;
  //@}            

  //@{
  //! Set year, month, day-in-month, hours, minutes or seconds.
  /*! Functions that set part of the Date object.
      Values that are too large will be converted, e.g.
      to e.g. next day. E.g.
      \code
      Date d("Mon Jan 01 23:00:00 CET 2001");
      d.hours(36);
      cout<< d.str();
      \endcode
      will print "Tue Jan 02 12:00:00 CET 2001".
  */
  void	year(unsigned short yr);
  void	month(unsigned short mo);
  void	day(unsigned short day);
  void	hours(unsigned short hrs);
  void	minutes(unsigned short mins);
  void	seconds(unsigned short secs);
  //@}            
  //! Set date to date with this yday in same year.
  void	yday(unsigned short yday);


  //! Returns number of secs since 1/1/1970.
  time_t 	time() const	{ return time_; }
  //! Set time.
  void		time(time_t time);

  //@{
  /*! Date comparison respects chronological order.
  */
  bool operator<(const Date& d) const	{ return (time() < d.time()); }
  bool operator==(const Date& d) const	{ return (time() == d.time()); }
  bool operator>(const Date& d) const	{ return d < *this; }
  bool operator!=(const Date& d) const	{ return ! (d == *this); }
  bool operator>=(const Date& d) const	{ return ! (*this < d); }
  bool operator<=(const Date& d) const	{ return ! (*this > d); }
  //@}

  //! Add a duration to a Date.
  Date& operator+=(const Duration&);
private:
  time_t		time_;	// nr of secs since Jan 1, 1970
};
}}

#endif
