#ifndef DVUTIL_DISPATCHER_H
#define DVUTIL_DISPATCHER_H
// $Id: dispatch.h,v 1.3 2003/08/11 09:38:12 dvermeir Exp $

namespace Dv {
namespace Util {
/** A class template framework to dispatch control based on some
 * value.
 * 
 * The framework assumes that there are two objects:
 *
 * @code
 * Input input; // represents input for application
 * Application application; // represents application
 * @endcode
 *
 * The Application class has a number of handler functions
 * @code
 * Application::handle_value_1(Input& input, const Value& value)
 * Application::handle_value_2(Input& input, const Value& value)
 * @endcode
 * one of which should be called. Which one to call depends
 * on a certain available value.
 *
 * The class template below makes it easy to dispatch to the correct
 * handler-function, based on the given value.
 * 
 * To use, an instance should define
 * @a Dispatcher<Application,Input,Value>::table as in the
 * example below.
 * @code
 * typedef std::string Value;
 * Dispatcher<Application,Input,Value>::Entry 
 * Dispatcher<Application,Input,Value>::table[] = {
 *   { "", &Application::default_handler },
 *   { "a", &Application::handle_a },
 *   { "b", &Application::handle_b },
 *   { "c", &Application::handle_c },
 *   { "", 0 } // sentinel
 * };
 * @endcode
 *
 * The Application class could then define an @a Application::main
 * function as follows.
 * @code
 * class Application {
 * public:
 *   bool main(Input& input, const Value& value) {
 *     return Dispatcher<Application,Input,Value>::handle(*this, input, value);
 *     }
 * };
 * @endcode
 * @sa Dv::Cgi::Cgi
 */
template <class Application, typename Input, typename Value>
class Dispatcher {
public:
  /** Handler type */
  typedef bool (Application::*Handler)(Input& input, const Value& value);
  /** Type of entry in dispatch table: if value is key, h will be
   * called */
  typedef struct { Value key; Handler h; } Entry;
  /** Actual dispatch table.
   * Define as in the example below.
   * @code
   * Dispatcher<Application,Input,std::string>::Entry 
   * Dispatcher<Application,Input,std::string>::table[] = {
   *   { "", &Application::default_handler },
   *   { "a", &Application::handle_a },
   *   { "b", &Application::handle_b },
   *   { "c", &Application::handle_c },
   *   { "", 0 } // sentinel
   * @endcode
   */
  static Entry table[];
  /** Main handler function. It will use @a value to dispatch a call to the
   * appropriate handler function.
   * An actual example can be found in @a test-dispatch.C in the
   * source directory.
   * @param  a application defining handler functions.
   * @param input object.
   * @param value used to select the handler function.
   */
  static bool handle(Application& a, Input& input,  const Value& value) {
    for (Entry* e(table); (e->h) ;  ++e)
      if (e->key==value)
        return (a.*(e->h))(input, value);
    return false;
    }
};

}}
#endif
