#ifndef DV_UTIL_ENUM2STR_H
#define DV_UTIL_ENUM2STR_H
// $Id: enum2str.h,v 1.5 2003/08/15 08:34:16 dvermeir Exp $
#include <stdexcept>
#include <dvutil/tostring.h>

namespace Dv {
namespace Util {
/** Convenient class template to support conversion of enum types
 * to/from strings.
 *
 * Example:
 * @code
 * const char* Dv::Util::enum_parser<X>::E_NAME("X"); // "name" of enum type X
 * Dv::Util::enum_parser<X>::enum_entry 
 * Dv::Util::enum_parser<X>::enum_table[] = {
 *   { A, "A" },
 *   { B, "B" },
 *   { C, "C" },
 *   { A, 0 } // should end with 0 in 2nd component of last array element!
 *   };
 *
 * std::cout << Dv::Util::enum2str(A) << std::endl;
 * std::cout << Dv::Util::str2enum<X>("A") << std::endl;
 * @endcode
 */
template<typename E>
class enum_parser {
public:
  /** Type of entry in enum_table. */
  typedef struct { E e; const char* s; } enum_entry;
  /** Table containing one enum_entry for each E value,
   * as well as a sentinel entry (with s=0).
   */
  static enum_entry enum_table[];
  /** Name of enum type, should be defined by user. */
  static const char*  E_NAME;
  /** Convert enum value to string.
   * @param e enum value
   * @return pointer to C-string representing e, never 0.
   * @exception std::logic_error if string representation not found.
   */
  static const char* enum2str(E e) throw (std::logic_error) {
    static const std::string NAME("enum2str");
    for (unsigned int i=0; enum_table[i].s; ++i)
      if (enum_table[i].e == e)
        return enum_table[i].s;
    throw std::logic_error(NAME + ": cannot convert " +
      E_NAME + " value " + tostring(static_cast<int>(e)));
    }
  /** Convert string to enum value.
   * @param s string to convert.
   * @return enum value corresponding to s.
   * @exception std::logic_error if s does not represent an E value.
   */
  static E str2enum(const std::string& s) throw (std::logic_error) {
    static const std::string NAME("str2enum");
    for (unsigned int i=0; enum_table[i].s ; ++i)
      if (enum_table[i].s == s)
        return enum_table[i].e;
    throw std::logic_error(NAME + ": cannot convert '" + s + "' to " + E_NAME);
    }
};

/** Convenience function to convert an enum value to a string.
 * @param e Enum value.
 * @return C-string representation of e (never 0).
 * @exception std::logic_error if string representation not found.
 */
template<typename E>
const char*
enum2str(E e) {
return enum_parser<E>::enum2str(e);
}

/** Convenience function to convert a string to an enum value.
 * @param s string representing an enum value.
 * @return enum value represented by string.
 * @exception std::logic_error if string does not represent an E value.
 */
template<typename E>
E
str2enum(const std::string& s) {
return enum_parser<E>::str2enum(s);
}

}}
#endif
