// $Id: fdstreambuf.C,v 1.15 2003/03/04 17:51:34 dvermeir Exp $

#include <config.h>

// sys/time.h is needed for timeval
#if TIME_WITH_SYS_TIME
# include <sys/time.h> // select
# include <time.h> 
#else
# if HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  include <time.h>
# endif
#endif

#include <iostream>
#include <sys/types.h> // time_t, size_t
#include <sys/stat.h>
#include <unistd.h>
#include <signal.h> // signal()
#include <assert.h>
#include <string.h> // strerror
#include <errno.h> // errno
#include <fcntl.h> // e.g O_RDONLY

#include "ntostr.h"
#include "fdstreambuf.h"

#include <string>
#include <dvutil/ntostr.h>

namespace Dv {
namespace Util {
using namespace std;
///////////////	constructor, destructor

fdstreambuf::fdstreambuf(int fdesc,size_t inbuf_sz, size_t outbuf_sz,
  time_t delay, bool noblocking, ostream* dbg): fd_(-1), delay_(delay), 
    non_blocking_(noblocking), inbuf_sz_(inbuf_sz>0?inbuf_sz:10), 
    outbuf_sz_(outbuf_sz), inbuf_(0), outbuf_(0), errno_(0), timedout_(false),
    debug_(dbg) {
// Initialize fd_.
fd(fdesc,non_blocking());

// Expect the system to set errno to EPIPE instead of throwing a
// SIGPIPE signal.
signal(SIGPIPE,SIG_IGN);

// Initialize input buffer, note that its size is at least 3.
inbuf_	= new char[inbuf_sz_];
setg(inbuf_,inbuf_+inbuf_sz_,inbuf_+inbuf_sz_);

// Initialize output buffer, note that its size may be 0.
if (outbuf_sz_) {
  outbuf_ = new char[outbuf_sz_];
  setp(outbuf_,outbuf_+outbuf_sz_);
  }
else
  setp(0,0);
}

fdstreambuf::~fdstreambuf() {
close();
if (inbuf_) delete [] inbuf_;
if (outbuf_) delete [] outbuf_;
}

///////////////	buffer management

int // return 0 if ok, -1 else
fdstreambuf::sync() { // flush pending output
if (outbuf_sz_==0) // unbuffered output
  return 0;
if (pbase()<pptr()) { // output buffer not empty
  int len = pptr()-pbase();
  int wlen = write(pbase(),len);
  setp(outbuf_,outbuf_+outbuf_sz_);
  if (wlen!=len)
    return -1;
  return 0;
  }
return 0; // nothing to flush
}

int
fdstreambuf::underflow() {
if (gptr()<egptr()) { // buffer not empty
  if (debug_)
    *debug_ << debug_header("underflow() =>")
            << *gptr() << " [from buffer]" << endl;
  return *gptr();
  }
assert(inbuf_sz_>0);
int n = read(inbuf_,inbuf_sz_);
if (n<=0) { // A bit naive: translate any serious error as EOF.
  if (debug_)
    *debug_ << debug_header("underflow() =>")
            << "=> " << EOF << " [eof]" << endl;
  return EOF;
  }
setg(inbuf_,inbuf_,inbuf_+n);
if (debug_)
  *debug_ << debug_header("underflow() =>")
  	  << *gptr() << " [after read of " << n << " chars]" << endl;
return *gptr();
}

int
fdstreambuf::overflow(int c) {
if (c==EOF) { // do not write EOF, just sync
  sync();
  return EOF;
  }
if (outbuf_sz_==0) { // unbuffered output
  char	buf[2];
  buf[0] = c;
  int wlen = write(buf,1);
  if (wlen!=1)
    return EOF;
  else
    return c;
  }
if (epptr()<=pptr()) { // buffer is ``full''
  if (sync()==-1)
    return EOF;
  }
*pptr() = c; // buffer not full
pbump(1);
return c;
}

///////////////	Blocking and waiting.

int
fdstreambuf::set_non_blocking() {
if (debug_) 
 *debug_ << debug_header("set_non_blocking");
int status(fcntl(fd_,F_SETFL,O_NONBLOCK));
if (status == -1)
  error(errno);
else
  non_blocking_ = true;
if (debug_) 
 *debug_ << "=> " << status << endl;
return status;
}

// return: 1 if ok to read/write on fd, 0 if timeout, <0 if error
int 
fdstreambuf::fdwait(int fd, time_t delay, int& syserrno, bool output) { 
if (fd<0) {
  syserrno = EBADF;
  return -1;
  }

fd_set	fds, exceptfds;
struct	timeval timeout_val;

FD_ZERO(&fds);
FD_SET(fd,&fds);
FD_ZERO(&exceptfds);
FD_SET(fd,&exceptfds);

timeout_val.tv_sec	= delay/1000;
timeout_val.tv_usec	= (delay%1000)*1000;

struct timeval*	timeval_ptr(delay ? (&timeout_val) : 0);

int nfds;
if (output)
  nfds = select(fd+1,NULL,&fds,&exceptfds,timeval_ptr);
else
  nfds = select(fd+1,&fds,NULL,&exceptfds,timeval_ptr);
if (nfds==0)
  syserrno = ETIMEDOUT;
else
  if (nfds<0)
    syserrno = errno;
return nfds;	
}


// return nr of ready-to-read fds: 1 if ok, 0 is timeout, <0 if error
int 
fdstreambuf::fdwait(bool output) const { 
int syserrno;
int err(fdwait(output,syserrno));
if (err<=0)
  error(syserrno);
if (debug_) 
  *debug_ << debug_header("fdwait()")
          << " => " << err << " [1=ok, 0 = timedout, -1 = error]" << endl;
return err;	
}

int // return nr of ready-to-read fds: 1 if ok, 0 is timeout, <0 if error
fdstreambuf::fdwait(bool output,int& syserrno) const { 
return fdwait(fd_, delay_ , syserrno, output);
}

///////////////	I/O

int
fdstreambuf::read(char *buf,size_t len) {
// Return
//	n>0	# of chars read
//	0	EOF
//	-1	timeout
//	-2	other error
string debug_msg;
if (debug_)
  debug_msg = debug_header("read") + "(" + ntostr(len) + "): ";
if (fd_<0) {
  if (debug_) 
    *debug_ << debug_msg << "fd_ = " << fd_ << " <0 => -2" << endl;
  return -2;
  }
int n;
do {
  int	nfds = iwait();
  if (nfds<0) { // iwait error
    if (debug_) 
      *debug_ << debug_msg << "iwait = " << nfds << " => -2" << endl;
    return -2;
    }
  if (nfds==0) { // timed out
    if (debug_) 
      *debug_ << debug_msg << "timedout => -1" << endl;
    error(ETIMEDOUT);
    return -1;
    }

  // Input is available.
  n = rread(buf,len); 

  if (debug_) {
    *debug_ << debug_msg << "read " << n << " chars: ";
    if (n>0) {
      *debug_ << "'";
      debug_->write(buf,n);
      *debug_ << "'" << endl;
      }
    else 
      *debug_ << " errno = " << errno << " (EINTR=" << EINTR << ")" << endl;
    }
  }
while (n==-1 && errno==EINTR);
// Read(2) returns -1 and sets errno=EINTR if ``a signal was caught
// during the read operation and no data was transferred''.
// We just redo the operation is this is the case, as wget-1.4.2 seems to do.
if (n==0) {
  if (debug_) 
    *debug_ << debug_msg << "read 0 chars => " << EOF << endl;
  return EOF;
  }
if (debug_) 
  *debug_ << debug_msg << "read " << n << " chars => " << n << endl;
return n;
}

int
fdstreambuf::rread(char *buf,size_t len) {
 // read(2) returns 0 on EOF, -1 if fd_<0, in the latter case, errno==EBADF
int n(::read(fd_,buf,len));
if (n<0)
  error(errno);
return n;
}

int
fdstreambuf::write(char *buf,size_t len) {
// Return
//	n>0	# of chars read
//	0	EOF
//	-1	timeout
//	-2	other error
string debug_msg;
if (debug_)
  debug_msg = debug_header("write") + "(" + ntostr(len) + "): ";
if (fd_<0) {
  if (debug_) 
    *debug_ << debug_msg << "fd_ = " << fd_ << " <0 => -2" << endl;
  return -2;
  }
int   m;
size_t   nleft =len;
char* pc = buf;
while (nleft>0) { 
  // if fd_<0, write() will return -1 and set errno==EBADF, we hope
  do {
    if (delay_>=0) {
      int	nfds = owait();
      if (nfds<0) { // owait error
        if (debug_) 
          *debug_ << debug_msg << "owait = " << nfds << " => -2" << endl;
        return -2;
        }
      if (nfds==0) { // timed out
	if (non_blocking()) {
	  // Try to write anyway, since select() on e.g. pipe may
	  // only succeed if kernel buffer is empty; i.e. it may
	  // fail while e.g. half the buffer is available for writing.
          m = rwrite(pc,nleft);
	  if ( (m == -1) && (errno != EINTR) ) { // real timeout
            error(ETIMEDOUT);
            if (debug_) 
              *debug_ << debug_msg << "timedout => -1" << endl;
	    return -1;
	    }
	  else {
	    error(0);
	    if (debug_)
	      *debug_ << debug_msg << "fake timeout, m = " << m << endl;
	    }
	  }
	else { // blocking, we accept the timeout
          error(ETIMEDOUT);
          if (debug_) 
            *debug_ << debug_msg << "timedout => -1" << endl;
	  return -1;
	  }
        }
      else 
        m = rwrite(pc,nleft);
      }
    else
      m = rwrite(pc,nleft);
    }
  while (m==-1 && errno==EINTR);
  if (m<=0) {
    if (debug_) 
      *debug_ << debug_msg << "rwrite fails => " << m << endl;
    return m;
    }
  nleft	-= m;
  pc	+= m;
  }
if (debug_) {
  *debug_ << debug_msg << "'";
  debug_->write(buf,len);
  *debug_ << "'" << endl;
  }
return len;
}

int
fdstreambuf::rwrite(char *buf,size_t len) {
// returns 0 on EOF, -1 if fd_<0, in the latter case, errno==EBADF
int n(::write(fd_,buf,len));
if (n<0)
  error(errno);
return n;
}

///////////////	fd stuff

bool
fdstreambuf::close() {
string debug_msg;
if (debug_)
  *debug_ << (debug_msg = debug_header("close")) << "(): start" << endl;
if (fd_>=0) {
  sync();
  // Upon successful completion, 0 is returned.
  // Otherwise, -1 is returned and errno is set to indicate the error.
  int status;
  do {
    status = ::close(fd_);
    }
  while (errno == EINTR);
  error(errno);
  if (debug_)
    *debug_ << debug_msg << "(): " << status << endl;
  timedout(false);
  fd_ = -1;
  return status==0;
  }
else
  if (debug_)
    *debug_ << debug_msg << "(): fd_ < 0" << endl;
return  true;
}

int
fdstreambuf::flags(std::ios::openmode mode) {
int f(0);
if (mode & ios::in)
  if (mode & ios::out)
    f |= O_RDWR;
  else
    f |= O_RDONLY;
if (mode & ios::out) {
  f |= O_CREAT;
  if (mode & ios::in)
    f |= O_RDWR; // bis repetita placet
  else
    f |= O_WRONLY;
  }
if (mode & ios::app)
  f |= O_APPEND;
if (mode & ios::trunc)
  f |= O_TRUNC;
return f;
}

// bool
// fdstreambuf::open(const char* fn, IOS_OPENMODE mode) {
// fd(::open(fn,flags(mode)));
// if (mode&ios::ate && !(mode&ios::app))
//   lseek(fd_,0,SEEK_END);
// if (fd_<0)
//   error(errno);
// return fd_>=0;
// }

bool
fdstreambuf::fd(int i, bool no_blocking) {
bool status(true);
if (fd_>=0)
  status = close();
fd_ = i;
if (no_blocking)
  set_non_blocking();
return status;
}

///////////////	error


void
fdstreambuf::error(int e) const {
if (debug_) 
  *debug_ << debug_header("error") << "(" << e << "), "
  	  << "current error = " << errno_ << " (" << strerror() << ")" << endl;
errno_ = e;
if (e==ETIMEDOUT)
  timedout(true);
else if (e == 0)
  timedout(false);
if (debug_) 
  *debug_ << debug_header("error") << "(" << e << ") "
  	  << "new error = " << strerror()
	  << endl;
}

string
fdstreambuf::strerror() const {
// cerr << "error = " << error() << endl;
char* msg = ::strerror(error());
return string(msg);
}

bool
fdstreambuf::timedout() const {
return timedout_;
}

bool
fdstreambuf::timedout(bool new_status) const {
if (new_status == false)
  if (timedout()) {
    timedout_ = false;
    if (error() == ETIMEDOUT) 
      error(0);
    return true;
    }
  else
    return false;
else  {
  if (timedout_ == true)
    return false;
  timedout_ = true;
  return true;
  }
}


// Utility private members.
string
fdstreambuf::debug_header(const string& funname) const {
static const string FDSTREAMBUF("fdstreambuf");
return FDSTREAMBUF + "[" + ntostr(fd()) + "]::" + funname;
}

}}
