#ifndef	DV_FDSTREAMBUF_H
#define	DV_FDSTREAMBUF_H
// $Id: fdstreambuf.h,v 1.21 2003/08/18 13:51:35 dvermeir Exp $

#include <iostream>
#include <string>
#include <sys/types.h> // for size_t, time_t
/** \file 
 *  The Dv::fdstreambuf class provides a streambuf specialization
 *  that manages a file descriptor (fd) and handles the possibility of
 *  timeout on input or open by failing an operation if it times out
*/
namespace Dv {
namespace Util {

/**
 * A streambuf specialization that manages a file descriptor and handles
 * timeouts.
 */
class fdstreambuf: public std::streambuf {
public:
  /**
   * The constructor simply allocates the buffers and stores the
   *  delay (in millisecs).
   *
   * \param fd filedescriptor to be used by this fdstreambuf
   * \param inbuf_sz size of input buffer.
   * \param outbuf_sz size of output buffer.
   * \param delay time, in millisecs, to wait before an open or input
   *   operation fails, a delay of 0 means ``wait foerver''.
   * \param non_blocking whether the underlying filedecriptor is in
   *   non-blocking (O_NONBLOCK) mode, see fctnl(2).
   *   Using a non-blocking streambuf prevents some blocking I/O
   *   operations. E.g. if select(2) indicates that writing is possible
   *   but the actual number of bytes to write is too large, the write may
   *   block indefinitely for a blocking Socket. For a non-blocking socket,
   *   the write(2) will return EAGAIN which will be handled
   *   appropriately by fdstreambuf.
   * \param debug pointer to ostream to write debug info on or 0, see
   *	fdstreambuf::debug().
   *
   *  The constructor also calls <code>signal(SIGPIPE,SIG_IGN)</code>,
   *  causing <code>SIGPIPE</code> signals to be ignored. Instead,
   *  the system should let the I/O operation fail and set errno to 
   *  <code>EPIPE</code>.
   */
  explicit fdstreambuf(int fd=-1,size_t inbuf_sz=1024,size_t outbuf_sz=1024,
      		       time_t delay=0, bool non_blocking = false,
		       std::ostream* debug = 0);
  /** 
   * This is already virtual in streambuf.
   */
  virtual ~fdstreambuf(); 

  /** \return size of input buffer */
  size_t inbuf_sz() const { return inbuf_sz_; }
  /** \return size of output buffer */
  size_t outbuf_sz() const { return outbuf_sz_; }
  /**
   * Return the number of millsecs to wait before an open or input operation
   * fails. A value of 0 means wait forever.
   *
   * \return The current delay value.
   */
  time_t  delay() const	{ return delay_; }

  /** 
   * Set delay (0 means infinite delay).
   *
   * \param d delay in millisecs, 0 means ``wait forever''.
   */
  void	  delay(time_t d) { delay_ = d; }

  /**
   * Return the blocking status.
   *
   * \return true iff the constructor was called with non_blocking or
   *	set_non_blocking() was called before.
   */
  bool  non_blocking() const	{ return non_blocking_; }

  /** 
   * Set non-blocking I/O.
   *
   * \return 0 upon success, -1 else in which case error() will
   * 	reflect the system errno.
   * \see fcntl(2)
   */
  int	  set_non_blocking();

  /**
   * Return file descriptor underlying this fdstreambuf object.
   *
   * \return The current value for the underlying file descriptor.
   */
  int	fd() const { return fd_; }

  /** 
   * Set the underlying file descriptor. The previous descriptor, if
   * valid, is closed.
   *
   * \param i new file descriptor.
   * \param non_blocking whether I/O should be non-blocking.
   * \return true iff close() of the previous valid file descriptor
   *         succeeded 
   * \see fdstreambuf::close()
   */
  bool fd(int i, bool non_blocking = false);

  /**
   * Close the underlying file descriptor.
   *
   * \return false iff close() of the underlying valid file descriptor
   *         failed.
   */
  virtual bool close();

  /**
   * Return true iff last I/O operation timed out.
   *
   * \return true iff the last I/O operation timed out or
   * 	timedout(true) was called since the last I/O operation.
   */
  bool	timedout() const;

  /** 
   * Set timedout() flag. This flag can only be cleared explicitly by
   * fdstreambuf::timedout(false) or by fdstreambuf::fd(int).
   *
   * @param timed_out New value for timedout().
   * @return true iff timedout status actually changed. It will
   *   return false, e.g. if the argument is false but the
   *   previous timedout status was not true.
   */
  bool	timedout(bool timed_out) const;

  /**
   * Return last relevant ::errno as set by the system.
   *
   * \return The last relevant errno as aset by the system.
   */
  int   error() const { return errno_; }

  /**
   * Return ::strerror(error()), a string representation of error().
   *
   * \return A string representation of error().
   */
  std::string strerror() const;

  /**
   * Set debug stream. If not null, trace info will be written to this
   * stream for all I/O operations.
   *
   * \param os Pointer to ostream or 0. In the latter case, no debug 
   *	information will be written.
   */
  void  debug(std::ostream* os) { debug_ = os; return; }

  /**
   * Return pointer to current debug stream.
   * The function may return 0, if debugging is ``off''.
   *
   * \return A pointer to the current debug stream or 0.
   *
   * Debug output may be generated for the following member functions:
   * set_non_blocking(), error(int), read(), write().
   */
  std::ostream*  debug() const { return debug_; }

  /** 
   * Check whether input or output on fd would block.
   *
   * \param fd file descriptor to check, none if -1
   * \param delay time (in millisecs) we are prepared to wait. If
   *   <code>0</code>, <code>fdwait()</code> may block indefinitely.
   * \param syserrno will contain system's errno if return -1
   * \param output false if fd is checked for input
   * \return <code>1</code> if input/output is available; <code>0</code>
   *   if a timeout occurred (in which case syserrno=ETIMEOUT)
   *   and a value <code><0</code> if another error occurred, in which case
   *   syserrno set to errno.
   *
   */
  static int fdwait(int fd, time_t delay, int& syserrno, bool output=false);

  /**
   * Override iostream::sync(), i.e. force all pending output to
   * be written using fdstreambuf::rwrite().
   *
   * \return 0 if ok, -1 else.
   */
  virtual int  sync();
protected:
  /**
   * Override iostream::underflow(), i.e. attempt to fill the input
   * buffer using fdstreambuf::rread().
   *
   * \return next input char or EOF.
   */
  virtual int  underflow();

  /**
   * Override iostream::overflow(). May call fdstreambuf::sync().
   *
   * \return c if the character was succeffuly written or EOF.
   */
  virtual int  overflow(int c);

  /**
   * Wait for I/O to become available. Wait for output or input,
   * depending on the value of the parameter. Sets error()
   * appropriately if return value is not 1.
   *
   * \param output Wait for output to be available if 
   * \return 1 if input is available, 0 if a timeout occurred; -1
   *  indicates an error condition (e.g. underlying fd not set).
   * \see fdwait
   */ 
  int  fdwait(bool output) const;

  /**
   * Wait for I/O to become available. Wait for output or input,
   * depending on the value of the parameter.
   *
   * \param output Wait for output to be available if 
   * \param syserrno System errorno (output variable).
   * \return 1 if input is available, 0 if a timeout occurred; -1
   *  indicates an error condition (e.g. underlying fd not set).
   * \see fdwait
   */ 
  int  fdwait(bool output, int& syserrno) const;

  /** 
   * Wait for input to become available. Equivalent to fdwait(0).
   *
   * \return 1 if input is available, 0 if a timeout occurred; -1
   *  indicates an error condition (e.g. underlying fd not set).
   */ 
  int  iwait() const { return fdwait(0); }

  /** 
   * Wait for output to become available. Equivalent to fdwait(0).
   *
   * \return 1 if input is available, 0 if a timeout occurred; -1
   *  indicates an error condition (e.g. underlying fd not set).
   */ 
  int  owait() const { return fdwait(1); }

  /** 
   * Attempt an input operation.
   *
   * \param buf Buffer where characters read will be put.
   * \param buf_sz Number of characters to read.
   * \return number of chars read, 0 on EOF, -1  if timeout, 
   *  	-2 if other error. 
   *
   * On timeout, error() is set to ETIMEDOUT. 
   * Input operations are retried if they were interrupted.
   *
   * \see error()
   */
  virtual int  read(char* buf,size_t buf_sz);

  /**
   * Attempt an output operation.
   * \param buf Buffer containing characters to write.
   * \param buf_sz Number of characters to write from buf.
   * \return number of chars written, 0 on EOF, -1  if timeout,
   *	-2 if other error.
   * On timeout, error() is set to ETIMEDOUT. 
   * Output operations are retried if they were interrupted.
   *
   * \see error()
   */
  virtual int  write(char* buf,size_t buf_sz);

  /**
   * Low level read function, default is ::read(fd(), buf, len);
   * may set error().
   * Derived classes can override this function.
   *
   * \param buf Buffer where characters read will be put.
   * \param len Number of characters to read.
   * \return number of chars read, 0 on EOF, -1 on error.
   *
   * \see error()
   */
  virtual int  rread(char* buf,size_t len);

  /**
   * Low level write function, default is ::write(fd(), buf, len); may
   * set error().
   * Derived classes can override this function.
   *
   * \param buf Buffer containing characters to write.
   * \param len Number of characters to write from buf.
   * \return number of chars read, 0 means no chars were written.
   *	On error, -1 is returned.
   *
   * \see error()
   */
  virtual int  rwrite(char* buf,size_t len);

  /**
   * Translate ios::openmode flags to open(2) flags like O_RDWR etc.
   *
   * \param mode Stream open mode.
   * \return A combination of O_RDWR, O_CREAT, O_APPEND, O_WRONLY,
   * 	O_TRUNC, O_RDONLY.
   *
   * Currently, the function understands combinations with
   * ios::in, ios::out, ios::app, ios::trunc.
   */
  static int  flags(std::ios::openmode mode);

  /** 
   * Set error status of fdstreambuf.
   *
   * \param e new error code.
   * 
   * \see fdstreambuf::error().
   */
  void error(int e) const;

private:
  fdstreambuf(const fdstreambuf&);
  fdstreambuf& operator=(const fdstreambuf&);
  /**
   * Underlying file descriptor, <0 if invalid.
   */
  int	       fd_;	
  /**
   * The number of millsecs to wait before an open or input operation
   * fails. A value of 0 means wait forever.
   */ 
  time_t       delay_; // timeout value, ==0 if infinite
  /**
   * Whether fd_ is non-blocking.
   */
  bool         non_blocking_;
  /**
   * Size of internal input buffer.
   */
  size_t       inbuf_sz_;
  /**
   * Size of internal output buffer.
   */
  size_t       outbuf_sz_;
  /**
   * Input buffer.
   */
  char*        inbuf_;
  /**
   * Output buffer.
   */
  char*        outbuf_;
  /**
   * Error status. A value of 0 means ok. Usually, the value of the
   * last ::errno as set by the system.
   */
  mutable int  errno_; 
  /**
   * Error status. A value of 0 means ok. Usually, the value of the
   * last ::errno as set by the system.
   */
  mutable bool timedout_; 
  /**
   * Pointer to stream to which debug info is to be written. If 0,
   * no debug information is generated.
   */
  std::ostream*     debug_;

  /**
   * Return string of the form ``fdstreambuf[fd]::function_name''
   *
   * \param funname Name to put after ``::''.
   * \return A string of the form ``fdstreambuf[fd]::function_name''
   * 	where fd is value of fd().
   */
  std::string debug_header(const std::string& funname) const;
};

}}
#endif
