// $Id: file.C,v 1.29 2003/09/19 14:04:31 dvermeir Exp $

#include  <pwd.h> // getpwnam
#include  <grp.h> // getgrnam
#include  <sys/types.h>  // stat(), opendir()
#include  <sys/stat.h>  // stat()
#include  <sys/param.h>  // MAXPATHLEN
#include  <sys/mman.h>  // mmap()
#include  <fcntl.h>  // O_RDONLY
#include  <stdio.h>  // open()
#include  <unistd.h>  // close(), rmdir(), getcwd()
#include  <dirent.h>  // opendir()
#include  <stdlib.h>  // free() mktemp() [ getcwd(0) ]
#include  <errno.h>  // errno
#include  <string.h>  // strerror()
#include  <utime.h>  // strerror()
// #include  <iomanip>  

#include <iostream>
#include <vector>
#include <algorithm>
#include <list>
#include <sstream>
#include  "file.h"
#include  "ntostr.h"
#include  "date.h"
#include  "systemerror.h"

namespace Dv {
namespace Util {
using namespace std;
/////////////////// FileError

const std::string FileError::NAME("Dv::Util::FileError");

/////////////////// File

File::File(const string& path) throw (FileError):path_(path), addr_(0) {
refresh();
}

File::File(const File& f) throw (FileError):path_(f.str()), addr_(0) {
refresh();
}

File&
File::operator=(const File& f) throw (FileError) {
if (this==&f)
  return *this;
unmap();
path_ = f.str();
refresh();
return *this;
}

File::~File() {
unmap();
}

File&
File::expand() throw (FileError) {
refresh();
path_ = realpath();
refresh();
return *this;
}

void
File::refresh() const throw (FileError) {
// set all data members:
//   stats_
//   exist_
//   type_
//   path_
//   map__

// set stats_, exist_
const_cast<File*>(this)->exist_ =  
	(lstat(path(),const_cast<struct stat*>(&stats_))==0);
if (!exist_) {
  // type_ = NONEXISTENT
  const_cast<File*>(this)->type_ = NONEXISTENT;
  // path_ remains unchanged
  // map_ = 0
  // don't know whether this makes sense, but it should do no harm
  const_cast<File*>(this)->unmap();
  return;
  }
// set type_
{
mode_t mode = stats_.st_mode;

if (S_ISLNK(mode))
  const_cast<File*>(this)->type_ =  SYMLINK;
else if (S_ISREG(mode))
  const_cast<File*>(this)->type_ =  REGULAR;
else if (S_ISDIR(mode))
  const_cast<File*>(this)->type_ =  DIRECTORY;
else if (S_ISCHR(mode))
  const_cast<File*>(this)->type_ =  SPECIALCHAR;
else if (S_ISBLK(mode))
  const_cast<File*>(this)->type_ =  SPECIALBLOCK;
else if (S_ISFIFO(mode))
  const_cast<File*>(this)->type_ =  FIFO;
else if (S_ISSOCK(mode))
  const_cast<File*>(this)->type_ =  SOCKET;
else const_cast<File*>(this)->type_ =  OTHER;
}
// (re)set path_
if (type_!=SYMLINK) // realpath for SYMLINK follows contents, which we do not want
  const_cast<File*>(this)->path_.assign(realpath());
}

std::string
File::absolute_path(const std::string& path) throw (Dv::Util::FileError) {
std::string resolved_path;
File::fullpath(path, resolved_path);
return resolved_path;
}

string
File::realpath() const throw (FileError) {
char    resolved_name[MAXPATHLEN];

string cwd(Directory::pwd());
if (::realpath(path(),resolved_name)==NULL) {
  char* msg = strerror(errno);
  if (chdir(cwd.c_str())==-1)
    throw FileError(path_+".realpath(): chdir failed");
  throw FileError(path_+".realpath() failed\n\tsystem msg = "+msg);
  }
return string(resolved_name);
}


time_t
File::last_modified() const throw (FileError) {
if (exists())
  return stats_.st_mtime;
return 0;
}

time_t
File::last_accessed() const throw (FileError) {
if (exists())
  return stats_.st_atime;
return 0;
}

size_t
File::size() const throw (FileError) {
if (exists())
  return stats_.st_size;
return 0;
}

mode_t
File::mode() const throw (FileError) {
if (exists())
  return stats_.st_mode & 07777;
return 0;
}

uid_t
File::owner() const throw (FileError) {
if (exists())
  return stats_.st_uid;
return 0;
}

gid_t
File::group() const throw (FileError) {
if (exists())
  return stats_.st_gid;
return 0;
}

File::Type
File::type() const throw (FileError) {
if (!exists())
  return NONEXISTENT;
else
  return type_;
}

bool 
operator==(const File& f1,const File& f2) {
//  2 Files are identical if they have the same i-node # on the same device.
f1.refresh();
if (!f1)
  return false;
f2.refresh();
if (!f2)
  return false;
return ((f1.stats_.st_ino==f2.stats_.st_ino) && (f1.stats_.st_dev==f2.stats_.st_dev));
}

bool 
operator<(const File& f1,const File& f2) {
//  2 Files are identical if they have the same i-node # on the same device.
f1.refresh();
if (!f1)
  return false;
f2.refresh();
if (!f2)
  return false;
if (f1.stats_.st_dev!=f2.stats_.st_dev)
  return (f1.stats_.st_dev < f2.stats_.st_dev);
else
  return (f1.stats_.st_ino < f2.stats_.st_ino);
}

const void*
File::map() throw (FileError) {
if (addr_)
  return addr_;
if (!exists())
  return 0;
int  fd = open(path(),O_RDONLY, S_IREAD|S_IRGRP|S_IROTH);
if (fd<0)
  return 0;
addr_ = mmap(static_cast<caddr_t>(0), size(), PROT_READ , MAP_PRIVATE, fd, 0);
close(fd);
if (addr_==reinterpret_cast<void*>(-1)) {
  addr_ = 0;
  return 0;
  }
else
  return addr_;
}

void
File::unmap() throw (FileError) {
if (addr_)
  if (munmap(static_cast<caddr_t>(addr_), size())<0)
    throw FileError(str()+".unmap(): munmap failed.");
addr_ = 0;
}

std::string&
File::content(std::string& s) const throw (FileError) {
if (! exists())
  throw FileError(str() + ": does not exist");
const void* pv(const_cast<Dv::Util::File*>(this)->map());
if (! pv)
  throw FileError(str() + ": cannot map");
s.append(static_cast<const char*>(pv), size());
return s;
}

std::string
File::content() const throw (FileError) {
std::string s;
return content(s);
}

bool
File::chown(uid_t uid) const throw (FileError) {
if (!exists())
  return false;
if (::chown(path(),uid,static_cast<gid_t>(-1))<0)
  return false;
return true;
}

bool
File::chgrp(const char* groupname) const throw (FileError) {
if (!exists())
  return false;
struct group* gr(0);
gr = getgrnam(groupname);
if (!gr)
  return false;
if (::chown(path(), static_cast<uid_t>(-1), gr->gr_gid)<0)
  return false;
return true;
}

bool
File::rm() throw (FileError) {
if (type_==File::DIRECTORY) 
  rmdir(path());
else
  unlink(path());
return !exists();
}

void
File::rmfr() throw (FileError) {
if (type_!=File::DIRECTORY) 
  rm();
else {
  Directory	dir(*this);
  const vector<string> files(dir.files());
  for (unsigned int i=0; i<files.size(); ++i) {
    File	file(dir.file(files[i]).c_str());
    file.rmfr();
    }
  rm();
  }
if (exists()) 
  throw FileError(str()+": rmfr fails");
}

bool
File::mv(const string& newpath) throw (FileError) {
if (exists()) {
  unmap();
  if (rename(path(),newpath.c_str())<0)
    return false;
  else {
    path_ = newpath;
    return exists(); // includes refresh()
    }
  }
else
  return false;
}

bool
File::touch(int mode) throw (FileError) {
int	fd;
if (exists()) {
  Date	now;
  struct utimbuf buf;
  buf.modtime = now.time();
  utime(path(), &buf);
  // fd = open(path(),O_WRONLY,mode);
  }
else {
  fd = open(path(),O_CREAT|O_WRONLY,mode);
  if (fd<0)
    return false;
  close(fd);
  }
return exists();
}

bool
File::mkdir(int mode) throw (FileError) {
if (exists())
  return false;
if (::mkdir(path(),mode)<0)
  return false;
return exists();
}

bool
File::mklink(const string& oldpath) throw (FileError) {
if (exists())
  return false;
if (::link(oldpath.c_str(),path())<0)
  return false;
return exists();
}

bool
File::mksymlink(const string& oldpath) throw (FileError) {
if (exists())
  return false;
// File	f(oldpath); // just to check that it exists
// if (f.exists()) // we use oldpath, _not_ f.path() which is absolute
  if (::symlink(oldpath.c_str(),path())<0)
    return false;
return exists();
}

bool
File::chmod(int mode) throw (FileError) {
if (!exists())
  return false;
if (::chmod(path(),mode)<0) 
  return false;
return true;
}

const string&
File::typestr(Type t) throw (FileError) {
static const string S_SYMLINK("SYMLINK");
static const string S_REGULAR("REGULAR");
static const string S_DIRECTORY("DIRECTORY");
static const string S_SPECIALCHAR("SPECIALCHAR");
static const string S_SPECIALBLOCK("SPECIALBLOCK");
static const string S_FIFO("FIFO");
static const string S_SOCKET("SOCKET");
static const string S_OTHER("OTHER");
static const string S_NONEXISTENT("NONEXISTENT");

if (t==SYMLINK) return S_SYMLINK;
else if (t==REGULAR) return S_REGULAR;
else if (t==DIRECTORY) return S_DIRECTORY;
else if (t==SPECIALCHAR) return S_SPECIALCHAR;
else if (t==SPECIALBLOCK) return S_SPECIALBLOCK;
else if (t==FIFO) return S_FIFO;
else if (t==SOCKET) return S_SOCKET;
else if (t==OTHER) return S_OTHER;
else if (t==NONEXISTENT) return S_NONEXISTENT;
throw FileError(ntostr(static_cast<int>(t))+" has no string representation");
}

static void
pathdirs(const string& path,vector<string>& dirs) {
string::size_type n0(0);
string::size_type n1(string::npos);
while ((n1=path.find_first_of('/',n0))!=string::npos) {
  dirs.push_back(path.substr(n0,n1-n0));
  n0 = n1+1;
  }
dirs.push_back(path.substr(n0));
}

string
File::relpath(const string& from) const throw (FileError) {
typedef vector<string>::const_iterator iter;
vector<string>	pfrom;
vector<string>	pto;
pathdirs(str(),pto);
pathdirs(from,pfrom);
pair<iter,iter>	e(mismatch(pfrom.begin(),pfrom.end(),pto.begin()));
ostringstream	oss;
for (iter i=e.first; i!=pfrom.end(); ++i)
  oss << "../";
for (iter i=e.second; i!=pto.end(); ++i) {
  if (i!=e.second)
    oss << "/";
  oss << *i;
  }
return oss.str();
}

string
File::relpath() const throw (FileError) {
return relpath(Directory::pwd());
}

/////////////////// Directory

Directory::Directory(const string& path) throw (FileError): File(path), files_() {
refresh();
}

Directory::Directory(const File& file) throw (FileError): File(file.path()), files_() {
refresh();
}

Directory::~Directory() {
}

string
Directory::file(const string& filename) const {
return str()+"/"+filename;
}

string
Directory::tempfile(const string& prefix) const throw (FileError) {
string tmpname = str() + "/" + prefix + "XXXXXX";
char tmp[tmpname.size()+1];
std::copy(tmpname.begin(), tmpname.end(), tmp);
tmp[tmpname.size()] = 0;
int fd(mkstemp(tmp));
if (fd==-1)
  throw FileError(string("tempfile(")+prefix+") failed");
::close(fd);
return tmp;
// book ok(fd!=-1);
// bool ok(mkstemp(tmp)!=-1);
// bool ok(mkstemp(const_cast<char*>(tmpname.c_str()))!=-1);
// if (ok) 
//  return tmpname;
// throw FileError(string("tempfile(")+prefix+") failed");
}

string
Directory::tempdir(const string& prefix) const throw (FileError) {
string tmpname = str() + "/" + prefix + "XXXXXX";
char tmp[tmpname.size()+1];
std::copy(tmpname.begin(), tmpname.end(), tmp);
tmp[tmpname.size()] = 0;
if (mkdtemp(tmp) == 0)
  throw FileError(SystemError().what());
return tmp;
}


void
Directory::refresh() const throw (FileError) {
File::refresh();
Directory* it(const_cast<Directory*>(this));
it->files_.clear();
if (exist_) {
  if (type_ != File::DIRECTORY)
    throw FileError(str()+" not a directory");
  DIR*	dir(opendir(path()));
  if (dir==0)
    throw FileError(string("opendir ")+path()+" failed");
  struct dirent* entry;
  while ((entry=::readdir(dir))) {
    string filename(entry->d_name);
    if ((filename!=".") && (filename!=".."))
      it->files_.push_back(entry->d_name);
    }
  ::closedir(dir);
  }
}

string
Directory::pwd() throw (FileError) {
static const string error_message(": getcwd failed in Directory::pwd()");
char    pathname[MAXPATHLEN];
char	*cwd(getcwd(pathname,MAXPATHLEN));
if (cwd==0) {
  char* sysmsg = strerror(errno);
  throw FileError(error_message+"\n\tsystem message: "+sysmsg);
  }
string	s(cwd);
return s;
}

bool
Directory::cd() const throw (FileError) {
if (!exists())
  return false;
return (chdir(path())!=-1);
}

string&
File::fullpath(string path, string& resolved_path) throw (FileError) {
if (path.size() == 0)
  throw FileError(path + " not a legal path");

if (path[0] != '/')
  path.insert(0, (Directory::pwd()+"/"));

// istringstream	iss(path.data(), path.size());
istringstream	iss(path);
list<string>	parts;

static const string dot(".");
static const string dotdot("..");

string part;
while (getline(iss, part,'/')) {
  switch (part.size()) {
    case 0:
      break;
    case 1:
      if (part!=dot)
        parts.push_back(part);
      break;
    case 2:
      if (part==dotdot) {
	if (parts.size())
          parts.pop_back();
	else
	  throw FileError(path + " no fullpath");
        }
      else
        parts.push_back(part);
      break;
    default:
      parts.push_back(part);
      break;
    }
  }

if (parts.size() == 0)
  resolved_path = "/";
else {
  resolved_path.resize(0);
  for (list<string>::const_iterator i=parts.begin(); i!= parts.end(); ++i)
    resolved_path.append("/").append(*i);
  }
return resolved_path;
}

string
File::fullpath() const throw (FileError) {
string resolved_path;
fullpath(path(), resolved_path);
return resolved_path;
}

}}
