#ifndef	DV_UTIL_FILE_H
#define	DV_UTIL_FILE_H
// $Id: file.h,v 1.28 2004/12/22 10:41:43 dvermeir Exp $
#include	<sys/types.h>
#include	<sys/stat.h>
#include	<string>
#include	<vector>
#include	<stdexcept>
#include	<iostream>


/** \file
 *
 * A \ref Dv::Util::File class object represents a pathname, possibly
 * associated with an existing file.
 *
 * Member functions are provided for retrieving or altering properties
 * such as existence, permissions, ownership etc.  In addition,
 * memory-mapping the contents of a File is also supported.
 * A  \ref Dv::Util::Directory class object represents a pathname,
 * possibly associated with an existing directory. It provides similar
 * functionality as File, augmented with the retrieval of the
 * directory's contents.
 */
namespace Dv {
namespace Util {
/**
 * A runtime exception thrown by some File operations.
 */
class FileError: public std::runtime_error {
public:
  /**
   * Name of class, is prepended to each message argument of the constructor.
   */
  static const std::string NAME;
  /**
   * Constructor, prepends NAME to message to obtain runtime_error::what().
   */
  FileError(const std::string& message): std::runtime_error(NAME+": "+message) {}
};

/**
 * A File objects represents a pathname, possibly an existing file.
 *
 * Example usage:
 *
 * \code
 * const char*	name;
 * File		f(name);
 * if (f.exists()) {
 *   Date		d(f.last_modified());
 * 
 *   cout << f.str() << endl;
 *   cout << std::string(d) << endl;
 * 
 *   const char*	pc = static_cast<const char*>(f.map());
 * 
 *   if (pc) // note: pc is not necessarily a C string
 *     for (unsigned int i=0;(i<f.size());++i,++pc)
 *       cout << *pc;
 *   }
 * else
 *   f.touch(); // make it exist
 * \endcode
 */

class File {
public:
  /**
   * Type of file, note NONEXISTENT for pathnames that do not refer to
   * a file.
   */
  enum Type { REGULAR, SYMLINK, DIRECTORY, SPECIALCHAR, 
	      SPECIALBLOCK, FIFO, SOCKET, OTHER, NONEXISTENT }; 
  /**
   * File permissions, you can use the | operator to combine them.
   */
  enum Mode { READ_OWNER = 0400,  WRITE_OWNER = 0200, EXEC_OWNER = 0100, 
  	      READ_GROUP = 040,   WRITE_GROUP = 020,  EXEC_GROUP = 010, 
  	      READ_OTHER = 04,    WRITE_OTHER = 02,   EXEC_OTHER = 01,
	      SETUID = 04000,     SETGID = 02000,
	      RWX_OWNER = 0700,   RWX_GROUP = 070,    RWX_OTHER = 07,
	      RW_OWNER = 0600,    RW_GROUP  = 060,    RW_OTHER = 06
	      };

  /**
   * Constructor, argument can be a relative or an absolute path.
   * \param path absolute or relative path
   */
  File(const std::string& path) throw (FileError);
  /**
   * Copy constructor.
   * \param f existing File object
   * \warning 
   * \code
   * File f(g);
   * \endcode
   * is equivalent with
   * \code
   * File f(g.str());
   * \endcode
   */
  File(const File& f) throw (FileError);
  /**
   * Assignment.
   * \param f existing File object
   * \warning 
   * \code
   * f = g;
   * \endcode
   * is, if \a &f!=&g , equivalent with
   * \code
   * File f(g.str());
   * \endcode
   */
  File& operator=(const File& f) throw (FileError);
  /**
   * Virtual destructor, calls unmap().
   */
  virtual ~File();
  /**
   * Change File object to a non-symbolic link file by
   * expanding symbolic links, if any.
   * \return reference to \a *this.
   * \exception FileError
   *   e.g. if there are too many symbolic links to follow.
   */
  File& expand() throw (FileError);

  /** Return string representation of File::Type. */
  static const std::string& typestr(Type) throw (FileError);
  /** Check existence of file. */
  bool        exists() const throw (FileError) { refresh(); return exist_; }
  /** Delegates to exists(). */
  operator    bool() const throw (FileError) { return exists(); }

  /** Returns a full, clean, absolute path as given by ::realpath(3C).
   * @warning For symbolic links the path used in the constructor is returned.
   */
  std::string str() const { return path_; }
  /** Returns a full, clean, absolute path as given by ::realpath(3C).
   * @warning For symbolic links the path used in the constructor is returned.
   */
  static std::string absolute_path(const std::string& path) throw (FileError);

  /** Delegates to str(). */
  operator const char*() const { return path_.c_str(); }
  /** 
   * Identical to str().c_str(). 
   * \return \a str().c_str()
   * \warning
   * For symbolic links the path used in the constructor is returned.
   */
  const char* path() const { return path_.c_str(); }
  /**
   * Retrieve full path without resolving symbolic links.
   * The result is absolute and does not contain '..' or '.'
   * components and no superfluous '/' chars.
   * \param path relative or absolute path
   * \param resolved_path (output)
   * \return reference to second parameter
   * \exception if the full path would climb up from \a '/',
   *  e.g. '/..' would throw an exception.
   * An empty \a path will also throw an exception.
   */
  static std::string& fullpath(std::string path, std::string& resolved_path) 
    throw (FileError);

  /**
   * Retrieve full path without resolving symbolic links.
   * The result is absolute and does not contain '..' or '.'
   * components and no superfluous '/' chars.
   * \return resolved path.
   * \exception if the full path would climb up from \a '/',
   *  e.g. '/..' would throw an exception.
   * An empty \a path will also throw an exception.
   */
  std::string fullpath() const throw (FileError);

  /** Only useful for symbolic links. */
  std::string realpath() const throw (FileError);
  /** 
   * Return path relative to from, which must be absolute. 
   * \param from path such that \a from/relpath(from)
   *   refers to this file
   * \return relative path \a x such that \a from/x refers to
   *   this file
   */
  std::string relpath(const std::string& from) const throw (FileError);
  /**
   * Return path relative to Directory::pwd() 
   * \return relpath(pwd())
   */
  std::string relpath() const throw (FileError);
  /** Return type of file (possibly File::NONEXISTENT). */
  Type	      type() const throw (FileError);
  /** Return true iff type() == File::DIRECTORY. */
  bool	      isdir() const throw (FileError) { return type() == DIRECTORY; }
  /** Return size in bytes of file, 0 if \a !exists(). */
  size_t      size() const throw (FileError);
  /** 
   * Return time of last modification of file, or 0 if it does not
   * exist.
   */
  time_t      last_modified() const throw (FileError);
  /**
   * Return time of last access to file, or 0 if it does not exist.
   */
  time_t      last_accessed() const throw (FileError);
  /**
   * Return permissions of file, \sa File::Mode.
   */
  mode_t      mode() const throw (FileError);
  /** Return uid of owner of file \sa Dv::Util::User. */
  uid_t       owner() const throw (FileError);
  /** Return gid of group-owner of file \sa Dv::Util::User. */
  gid_t       group() const throw (FileError);

  /**
   * Change the ownership of an existing file, return true iff successful.
   * \param uid id of new owner.
   * \return true iff new owner is \a uid.
   */
  bool	      chown(uid_t uid) const throw (FileError);
  /**
   * Change group-ownership of an existing file, return true iff successful.
   * \param groupname of new group owner.
   * \return true iff new group owner is \a groupname.
   */
  bool	      chgrp(const char* groupname) const throw (FileError);
  /**
   * Change permissions on an existing file.
   * \param mode representing permissions.
   * \return true iff permissions where succesfully set.
   * \sa File::Mode
   */
  bool        chmod(int mode=0755) throw (FileError);
  /**
   * Remove a file, for directories, rmdir is used.
   * \return true if \a exists() false afterwards.
   */
  bool	      rm() throw (FileError); 
  /** Like rm -fr. Throws FileError if exists() afterwards. */
  void	      rmfr() throw (FileError); 
  /** 
   * Rename file, return true iff succeeds, also renames *this.
   * \param newpath new name of file.
   * \return true iff rename succeeded.
   */
  bool	      mv(const std::string& newpath) throw (FileError); 
  /** Creates or updates modification time of file. */
  bool	      touch(int mode=0644) throw (FileError); 
  /** Create a directory str(), precondition is !exists(). */
  bool 	      mkdir(int mode=0755) throw (FileError);
  /** Create a hard link to path, precondition is !exists(). */
  bool 	      mklink(const std::string& path) throw (FileError);
  /**
   * Create a symbolic link to path.
   * \return false if the file already exists, true if 
   *   the link was succesfully created.
   * \pre exists() returns false
   * \warning
   *   There is no check on the path parameter.
   */
  bool 	      mksymlink(const std::string& path) throw (FileError);

  /**
   * Compare File objects.
   *
   * File comparison is independent of the name, the operator
   * compares i-nodes and devices. Nonexistent File objects
   * are never equal.
   * \warning This function does not expand symbolic links.
   * Thus, if \a s is a symbolic link to \a f, \a s and 
   * \a f will be different.
   */
  friend bool 	operator==(const File&,const File&);
  /**
   * Compare File objects.
   *
   * File comparison is independent of the name, the operator
   * compares i-nodes and devices. Nonexistent File objects
   * are never equal.
   * \warning This function does not expand symbolic links.
   * Thus, if \a s is a symbolic link to \a f, \a s and 
   * \a f will be different.
   */
  friend bool 	operator!=(const File& f1,const File& f2) { return ! (f1==f2); }
  /**
   * Compare File objects.
   * 
   * File comparison is independent of the name, the operator
   * compares i-nodes and devices. Nonexistent File objects
   * are never equal.
   * \warning This function does not expand symbolic links.
   * Thus, if \a s is a symbolic link to \a f, \a s and 
   * \a f will be different.
   */
  friend bool 	operator<(const File&,const File&);

  /**
   * Map the contents of an existing file into memory. Return 0 upon failure.
   * Map() returns a pointer to the contents of the file, mapped into
   * virtual memory
   * \warning the returned pointer is not a C string, use size() to
   *  obtain the size of the map.
   */
  const void*	map() throw (FileError);
  /**
   * Release a mapping, if any, obtained by File::map().
   * Unmapping a file that is not mapped is a noop.
   * File::unmap() is automatically called by the destructor.
   * \exception throws a FileError iff munmap(2) failed.
   */
  void		unmap() throw (FileError);
  /** @return address of map for mapped file, or 0 otherwise. */
  const void*	mapped() const { return addr_; }

  /** Append contents of file to string. 
   * @param s string to append to
   * @return reference to s
   */
  std::string& content(std::string& s) const throw (FileError);

  /** Return contents of file to string. 
   * @return contents of file
   */
  std::string content() const throw (FileError);

protected:
  /** Updates private data members, exist_ and type_. */
  virtual void	refresh() const throw (FileError);
  /** True iff there is a file with pathname str(). */
  bool		exist_; // updated by refresh()
  /** Type of file. */
  Type		type_;
private:
  struct stat	stats_; // updated by refresh()
  std::string	path_;
  void*		addr_;
};

/**
 * A Directory represents a pathname that refers to a possibly
 * non-existing directory.  It is privately derived from File,
 * relevant File members are available,
 * in addition tot the files() member that retrieves the files in an
 * existing directory.
 *
 * \par 
 * The following File members are available also in Directory:
 *
 *  File::exists,
 *  File::mkdir,
 *  File::path,
 *  File::str,
 *  File::realpath,
 *  File::relpath,
 *  File::last_modified,
 *  File::last_accessed,
 *  File::touch,
 *  File::chown,
 *  File::chgrp,
 *  File::chmod,
 *  File::rm,
 *  File::mv,
 *  File::rmfr
 */
class Directory: private File {
public: // a directory always exists
  /** Constructor, note that path may not exist. */
  Directory(const std::string& path) throw (FileError); 
  /** Equivalent to Directory(file.str()); */
  Directory(const File& file) throw (FileError);
  /** Destructor. */
  ~Directory();

  /**
   * Return files (only filenames, not absolute paths) in the directory.
   * If the directory does not exist, the vector is empty. 
   * The directory entries ``.'' and ``..'' are not included
   * in the output of Directory::files().
   */
  const std::vector<std::string>& files() const throw (FileError) { 
   			  refresh(); return files_; }
  /**
   * Return str()+"/"+filename, i.e. a pathname for a filename in directory.
   */
  std::string file(const std::string& filename) const;
  /** 
   * Return \a str()+"/"+temp_filename, where \a temp_filename starts
   * with \a prefix, i.e. a pathname for a temporary file in the directory.
   * @param prefix of name of file to be created in this directory.
   * @return the filename
   * \warning This function calls mkstemp() which will actually create
   *   the corresponding file.
   */
  std::string tempfile(const std::string& prefix) const throw (FileError);

  /** 
   * Return @a str()+"/"+temp_dirname, where \a temp_dirname starts
   * with @a prefix, i.e. a pathname for a temporary subdirectory in the directory.
   * @param prefix of name of file to be created in this directory.
   * @return the directoryname
   * @warning This function calls mkdtemp() which will actually create
   *   the corresponding directory.
   */
  std::string tempdir(const std::string& prefix) const throw (FileError);

  /** Change working directory. */
  bool cd() const throw (FileError);
  /** Return absolute pathname of current working directory. */
  static std::string pwd() throw (FileError);

  /** \sa File::exists */
  using File::exists;
  /** \sa File::mkdir */
  using File::mkdir;
  /** \sa File::path */
  using File::path;
  /** \sa File::str */
  using File::str;
  /** \sa File::realpath */
  using File::realpath;
  /** \sa File::relpath */
  using File::relpath;
  /** \sa File::last_modified */
  using File::last_modified;
  /** \sa File::last_accessed */
  using File::last_accessed;
  /** \sa File::touch */
  using File::touch;
  /** \sa File::chown */
  using File::chown;
  /** \sa File::chgrp */
  using	File::chgrp;
  /** \sa File::chmod */
  using File::chmod;
  /** \sa File::rm */
  using File::rm;
  /** \sa File::mv */
  using File::mv;
  /** \sa File::rmfr */
  using File::rmfr;


private:
  void	refresh() const throw (FileError);
  void	init() const throw (FileError);

  std::vector<std::string> files_;
};
}}
#endif
