#ifndef	DV_UTIL_FILTERSTREAMBUF_H
#define	DV_UTIL_FILTERSTREAMBUF_H
// $Id: filterstreambuf.h,v 1.6 2003/08/05 13:02:07 dvermeir Exp $

#include	<iostream>
// #include	<streambuf>

/** \file 
 * The Dv::Util::filterstreambuf class is a template class that
 * is used to define Dv::Util::filterstream objects.
 */
namespace Dv {
namespace Util {
/**
 * A streambuf specialization that cooperates with a filter class object.
 * A filterstreambuf does not buffer data, but immediately passes them on
 * (resp. obtains them from) to the filter.
 *
 * The requirements for the Filter template parameter are as follows:
 *
 *  \code
 *  class Filter {
 *  public:
 *    int get(); // return next input char or Traits::eof
 *    int put(int c); // output c if c!=EOF, return EOF if error
 *    int sync(); // synchronize filter, return 0 (ok) or -1 (EOF)
 *    void close(); // close filter, no obligations..
 *    ios_base::iostate state(); // return status of filter
 *  };
 *  \endcode
 * Here, \a state() is not used by filterstreambuf but may be
 * convenient for streams based on a filterstreambuf: the
 * constructor can use basic_ios::clear(filter.state()) to inherit
 * the initial state of the filter.
 * 
 *  \see Dv::Util::teestream
 */
template <class Filter>
class filterstreambuf: public std::streambuf {
public:
  /**
   * The constructor remembers the filter to which I/O will be forwarded.
   * \param filter to which I/O operations will be forwarded.
   * \warning The filter is not copied, the management of the filter
   *   is not the filterstreambuf's responsibility (only a reference
   *   to the filter is kept).
   */
  explicit filterstreambuf(Filter& filter): filter_(filter) {
    setg(inbuf_,inbuf_+1,inbuf_+1); // very small input buffer
    setp(outbuf_,outbuf_); // no output buffer
    }
  /**
   * This is already virtual in streambuf.
   */
  virtual ~filterstreambuf() {}
  /**
   * \return underlying filter (const version).
   */
  const Filter&	filter() const { return filter_; }
  /**
   * \return underlying filter (non-const version).
   */
  Filter&	filter() { return filter_; }

  /**
   * Set underlying filter.
   * \param filter to which I/O operations will be forwarded.
   * \warning The filter is not copied, the management of the filter
   *   is not the filterstreambuf's responsibility. 
   */
  filterstreambuf& filter(Filter& filter) { filter_ = filter; return *this; }

  /**
   * Close -- indicate intention not to use the filter anymore.
   * \return return status of Dv::Util::filterstreambuf::sync.
   */
  bool close() { int tmp=sync(); filter_.close(); return tmp; }

  /**
   * Override iostream::sync().
   * \return return status of \a filter_.sync() .
   */
  virtual int  sync() { return filter_.sync(); }

protected:
  /**
   * Override iostream::underflow().
   * \return next input char or EOF
   */
  virtual int  underflow() { 
    if (gptr()<egptr()) // buffer not empty
      return *gptr();
    if ( (inbuf_[0] = filter_.get()) != EOF)
      setg(inbuf_,inbuf_,inbuf_+1);
    return inbuf_[0];
    }  
  /**
   * Override iostream::overflow().
   * \param c next char to put
   * \return status of filter_.put(c)
   */
  virtual int  overflow(int c) { return filter_.put(c); }

private:
  Filter&	filter_;
  char		inbuf_[1];
  char		outbuf_[1]; // unused
};

}}
#endif
