#ifndef DVUTIL_INCLUDESTREAM_H
#define DVUTIL_INCLUDESTREAM_H
// $Id: includestream.h,v 1.7 2003/07/14 12:12:09 dvermeir Exp $

#include <string>
#include <list>
#include <iostream>
#include "filterstreambuf.h"

namespace Dv {
namespace Util {
/**
 * @file
 * This files defines a class includestream, which uses
 * a filterstreambuf to process @a #include @a "filename" directives.
 */

/**
 * A stream that can include other files.
 * An includestream is an @a istream that replaces lines
 * of the form
 * @code
 * #include "file"
 * #include <file>
 * @endcode
 * by the contents of the @c file. If @c file is between
 * double quotes and is not absolute, it is interpreted as relative
 * to the current directory. Otherwise (i.e. if it is of
 * the form @c &lt;file&gt; , it will be searched for (if it is
 * not absolute) in a list of directories.
 * @warning
 * <ol>
 * <li>An exception (@a runtime_error ) is thrown upon an
 *  include error (syntax, file not found, ..)
 * <li>No check is made (yet) for recursive include directives.
 * </ol>
 */
class includestream: public std::istream {
public:
  /** Type of optional second argument: a list of directories
   * to search for included files.
   */
  typedef std::list<std::string> directories;
private:
  /**
   * Implements Filter interface for filterstreambuf
   */
  class IncludeFilter {
  public:
    /** Constructor.
     * @param is underlying input stream (need not be a file stream).
     * @param dirs pointer (may be 0) to a list of directories
     *  to search if an include directive is of the @c &lt;file&gt; form.
     */
    IncludeFilter(std::istream& is, directories* dirs);
    /** Constructor.
     * @param is underlying input stream (need not be a file stream).
     * @param dir a single directory
     *  to search if an include directive is of the @a &lt;file&gt; form.
     */
    IncludeFilter(std::istream& is, const std::string& dir);
    /** Destructor. */
    ~IncludeFilter();
    int put(int c) { return EOF; }
    int get();
    int sync() { return -1; }
    void close() {}
    std::ios::iostate state() const;
    /** @return directories that will be searched when processing
     * include directives of the form @c &lt;file&gt;.
     */
    const directories& dirs() const { return dirs_; }
  private:
    /** Underlying istream from where the filter takes its input. */
    directories dirs_;
    /** Current input line. */
    std::string line_;
    /** Current position in input line, std::string::npost if none. */
    std::string::size_type pos_;
    /** Stack of open files. */
    std::list<std::istream*> is_;
    /** Auxiliary function: pop @a &is from @a dirs_ . 
     * @param is stream to pop.
     */
    void pop(std::istream& is);
  };
public:
  /**
   * Uses open stream.
   * @param stream from which original input is taken
   * @param dirs pointer (may be 0) to a list of directories
   *   to search if an include directive is of the @c &lt;file&gt; form.
   * @warning No check for recursive include is made (yet).
   */
  includestream(std::istream& stream, directories* dirs = 0);
  /**
   * Uses open stream.
   * @param stream from which original input is taken
   * @param dir a single directory
   *  to search if an include directive is of the @a &lt;file&gt; form.
   * @warning No check for recursive include is made (yet).
   */
  includestream(std::istream& stream, const std::string& dir);
  /** Destructor. */
  ~includestream();

  /** Try to find a file @a d/filename for some directory @a d
   * in @a dirs. Return zero-length string if not found. The
   * directories are tried in order.
   * @param filename (relative) path of filename. If filename
   *   is absolute, i.e. starts with @a / , only its existence
   *   (without considering @a dirs ) is checked.
   * @param dirs list of directories to try
   * @return full pathname of first existing file @a d/filename 
   *  for some @a d in @a dirs, or zero-length string. If @a filename
   *  starts with @a '/' , only the existence of @a filename is
   *  checked.
   */
  static std::string find(const std::string& filename, const directories& dirs);
private:
  IncludeFilter	filter_;
};
}}

#endif
