#ifndef DV_UTIL_LOGSTREAM_H
#define DV_UTIL_LOGSTREAM_H
// $Id: logstream.h,v 1.11 2004/12/22 10:10:41 dvermeir Exp $

#include <string>
#include <iostream>
#include "filterstreambuf.h"

/** \file
 *  This file defines a class logstream, which uses
 *  a Dv::Util::filterstreambuf to prepend 
 *  a date and a prefix string to each output line.
 */

namespace Dv {
namespace Util {

/** 
 * A logstream is an ostream useful for logging.
 * Each line is preceeded by the date and an
 * application-determined identification string.
 * @warning
 *  Output is line-oriented and buffered, i.e.
 *  output will be kept until @a '\\n' is seen, at which time
 *  the whole line (preceeded by prefix and date information)
 *  will be written to the output.
 *  
 */
class logstream: public std::ostream {
private:
  /**
   * Implements Filter interface for filterstreambuf
   */
  class LogFilter {
  public:
    /**
     * Constructor.
     * \param path name of file on which log will be written.
     * \param prefix: each output line will start with the
     *   current date and the prefix string \sa Dv::Util::Date
     */
    LogFilter(const std::string& path,const std::string& prefix);
    /**
     * Constructor, stream is managed by client.
     * \param stream open output stream on which log will be written.
     * \param prefix: each output line will start with the
     *   current date and the prefix string \sa Dv::Util::Date
     */
    LogFilter(std::ostream& stream,const std::string& prefix);
    /**
     * Destructor.
     */
    ~LogFilter();
    /** Output a character.
     * @param c char to output
     * @return @a c is all went well, @a EOF otherwise.
     */
    int put(int c);
    /** Input a character, always fails.
     * @return EOF, this is an output-only filter.  */
    int get() { return EOF; }
    /** Flush buffers.
     * @return 0 if ok, @a EOF otherwise
     */
    int sync();
    /**
     * This is a noop.
     */
    void close() {}
    /**
     * Set prefix that will be printer at the start of each output line.
     */
    void prefix(const std::string& prefix) { prefix_ = prefix; }
    /**
     * \return prefix that is used on output 
     * \sa prefix(const std::string&)
     */
    const std::string& prefix() const { return prefix_; }
    /**
     * \return iostate of underlying stream
     */
    std::ios::iostate state() const;
  private:
    bool owner_;
    std::string buf_;
    std::string prefix_;
    std::ostream* stream_;
    bool lastnl_;
  };
public:
  /**
   * Creates ofstream for path.
   * The stream will be opened with mode \a ios:app, i.e.
   * a seek to the end of the stream will be performed
   * before each output operation.
   * \param path on which log output will be written.
   * \param prefix each output line will start with the
   *  date and this prefix string
   * \sa Dv::Util::Date
   */
  logstream(const std::string& path, const std::string& prefix);
  /**
   * Constructor.
   * \param stream existing stream on which output will be written.
   * \param prefix each output line will start with the
   *  date and this prefix string
   */
  logstream(std::ostream& stream, const std::string& prefix);
  /**
   * Destructor.
   */
  ~logstream();
  /**
   * \return current prefix that is put at the start of each line.
   * \sa Dv::Util::logstream::logstream
   */
  const std::string& prefix() const { return filter_.prefix(); }
  /**
   * \param s prefix that is put at the start of each line.
   * \sa Dv::Util::logstream::logstream
   */
  void prefix(const std::string& s) { filter_.prefix(s); }
private:
  LogFilter	filter_;
};

}}
#endif
