#ifndef DV_NTOSTR_H
#define DV_NTOSTR_H
// $Id: ntostr.h,v 1.19 2003/08/18 13:56:52 dvermeir Exp $

#include <string>
#include <stdexcept>

namespace Dv {
namespace Util {
/** Produces a string representation of its int argument.
 * The representation is determined by a sprintf format string.
 * @param i integer to represent
 * @param fmt format C-string, the default (if 0) is <code>"%d"</code>.
 * @return string representation of @a i, according to @a fmt.
 * @warning The buffer size is 20.
 */
std::string ntostr(int i, const char* fmt=0);

/** Produces a string representation of its unsigned int argument.
 * The representation is determined by a sprintf format string.
 * @param u unsigned integer to represent
 * @param fmt format string, the default (if 0) is <code>"%u"</code>.
 * @return string representation of @a u, according to @a fmt.
 * @warning The buffer size is 20.
 */
std::string ntostr(unsigned int u, const char* fmt=0);

/** Produces a string representation of its long argument.
 * The representation is determined by a sprintf format string.
 * @param l long integer to represent
 * @param fmt format string, the default (if 0) is <code>"%ld"</code>.
 * @return string representation of @a l, according to @a fmt.
 * @warning The buffer size is 20.
 */
std::string ntostr(long l, const char* fmt=0);

/** Produces a string representation of its unsigned long argument.
 * The representation is determined by a sprintf format string.
 * @param ul unsigned long integer to represent
 * @param fmt format string, the default (if 0) is <code>"%lu"</code>.
 * @return string representation of @a ul, according to @a fmt.
 * @warning The buffer size is 20.
 */
std::string ntostr(unsigned long ul, const char* fmt=0);

/** Produces a string representation of its double argument.
 * The representation is determined by a sprintf format string.
 * @param d unsigned long integer to represent
 * @param precision number of digits after decimal point in representation
 * @param fmt format string, the default (if 0) is <code>"%.Pf"</code>.
 * @return string representation of @a d, according to @a fmt and precision.
 * @warning The buffer size is 20.
 */
std::string ntostr(double d, unsigned int precision = 7, const char* fmt=0);

/** Try to convert the initial part of a C string to a long.
 * @param p C-string to convert
 * @param l long receiving the value
 * @param base to be used in the conversion. The default base (0) reads
 * decimal or octal (if the string starts with 0) numbers.
 * @return true iff an initial part of the C string p can be converted to a long value.
 * @warning If decimal is necessary, specify base 10 explicitely!
 */
bool cstr2long(const char* p,long& l,int base=0);

/** Try to convert the initial part of a C string to an unsigned long.
 * @param p C-string to convert
 * @param ul unsigned long receiving the value
 * @param base to be used in the conversion. The default base (0) reads
 * decimal or octal (if the string starts with 0) numbers.
 * @return true iff an initial part of the C string p can be converted to an unsigned long value.
 * @warning If decimal is necessary, specify base 10 explicitely!
 */
bool cstr2ulong(const char* p, unsigned long& ul, int base=0);

/** Try to convert the initial part of a C string to a double.
 * @param p C-string to convert
 * @param d double receiving the value
 * @return true iff an initial part of the C string p can be converted
 * to a double value.
 */
bool cstr2double(const char* p, double& d);

/** Change (in-place) all chars in string to lower case. 
 * @param s string to convert.
 * @return reference to first argument.
 */
std::string& s2lower(std::string& s);

/** Change (in-place) all chars in string to upper case. 
 * @param s string to convert.
 * @return reference to first argument.
 */
std::string& s2upper(std::string& s);

/** Remove leading and trailing white noise (' ', '\\n', '\\t') from string.
 * The argument is changed and returned
 * @param s string to trim
 * @return reference to the argument
 */
std::string& strim(std::string& s);

/** Remove leading and trailing white noise (' ','\\n','\\t') from C string.
 * @param pc C-string to trim.
 * @return pointer to first non-white-space character in the argument.
 * @warning The argument is changed. Since the return value points to the
 *  first non-white-space char in the argument, the following is
 *  a recepy for disaster:
 *  @code
 *  char* pc = new char[10];
 *  ..
 *  pc = strim(pc);
 *  delete[] pc;
 *  @endcode
 */
char* strim(char* pc);

/** Class wrapper for int.
 * This class is useful for automatic conversion of a string to an int.
 *  @code
 *  int i = Int("14");
 *  @endcode
*/
class Int {
public:
  /** Constructor.
   * @param s containing string representation of integer.
   * @param base used when converting s. The default is 0, in which case the function
   *   reads decimal or octal (if the string starts with 0) numbers.
   * @warning If decimal is necessary, specify base 10 explicitely!
   */
  Int(const std::string& s,int base=0) throw (std::domain_error);
  /** Conversion to int. */
  operator int() const { return int_; }
  /** Return associated numeric value. */
  int value() const { return int_; }
private:
  int  int_;
};

/** Class wrapper for long.
 * This class is useful for automatic conversion of a string to an long.
 *  @code
 *  long l = Long("14909090");
 *  @endcode
 */
class Long {
public:
  /** Constructor.
   * @param s containing string representation of integer.
   * @param base used to convert @a s. The default is 0, in which case the function
   *   reads decimal or octal (if the string starts with 0) numbers.
   * @warning If decimal is necessary, specify base 10 explicitely!
   */
  Long(const std::string& s,int base=0) throw (std::domain_error);
  /** Conversion to int. */
  operator long() const { return long_; }
  /** Return associated numeric value. */
  long value() const { return long_; }
private:
  long  long_;
};

/** Class wrapper for double.
 *  This class is useful for automatic conversion of a string to an
 *  double.
 *  @code
 *  double d = Double("3.14");
 *  @endcode
 */
class Double {
public:
  /** Constructor.
   * @param s containing string representation of integer.
   */
  Double(const std::string& s) throw (std::domain_error);
  /** Conversion to double. */
  operator double() const { return double_; }
  /** Return associated numeric value. */
  double value() const { return double_; }
private:
  double  double_;
};

}}
#endif
