#ifndef DV_UTIL_PERIOD
#define DV_UTIL_PERIOD
// $Id: period.h,v 1.4 2002/10/18 11:41:26 dvermeir Exp $

#include <stdexcept>
#include <dvutil/date.h>

namespace Dv {
namespace Util {

//! Class representing a duration.
/*! This is intended for administrative applications, e.g. adding
    a duration of 24 hrs to a Date will result in the next day,
    same time, even if a daylight savings time switch happens in
    the period.

    There are 3 incompatible units:
    <ul>
    <li> seconds (and minutes, hours, days, weeks).
    <li> months (same reasoning as above).
    <li> years (idem)
    </ul>

*/
class Duration {
public:
  //! Rather inconvenient, the static functions below are probably more useful.
  Duration(size_t years, size_t months, size_t seconds): years_(years),
	months_(months), seconds_(seconds) {}

  //@{
  /*! Convenience replacements for constructor.
  */
  static Duration mins(size_t mins, size_t secs = 0);
  static Duration hrs(size_t hrs, size_t mins=0, size_t secs = 0);
  static Duration days(size_t days, size_t hrs=0, size_t mins=0, size_t secs = 0);
  static Duration weeks(size_t wks, size_t days=0, size_t hrs=0, 
			size_t mins=0, size_t secs = 0);
  static Duration months(size_t months);
  static Duration years(size_t years, size_t months=0);
  //@}

  //! Add two duration: separately adds the differen second/month/year units.
  Duration operator+(const Duration& d) const;
  //! Add a duration to a duration: separately adds the differen second/month/year units.
  Duration operator+=(const Duration& d);

  //! Number of years of duration.
  size_t	years() const { return years_; }
  //! Number of months of duration.
  size_t	months() const { return months_; }
  //! Number of seconds of duration.
  size_t	seconds() const { return seconds_; }

  //! Return string representation of duration. 
  /*! This has the form
      \code
      YY-MM-DD HH:MM:SS
      \endcode
  */
  std::string	str() const;
private:
  size_t	years_;
  size_t	months_;
  unsigned long	seconds_;
};

//! A Period is simply a Date with a Duration.
class Period {
public:
  //! Constructor.
  Period(const Date& start, const Duration& size): start_(start), size_(size) {}
  //! Constructor, the duration will be such that <code>start + size = end</code>.
  Period(Date start, Date end) throw (std::runtime_error);

  //! Start date of period.
  Date	   	  start() const { return start_; }
  //! End date of period, i.e. <code>start + size</code>.
  Date	   	  end() const;
  //! Size of period.
  const Duration& size() const { return size_; }
private:
  Date	   start_;
  Duration size_;
};

}}

//! Write <code>d.str()</code>.
std::ostream&
operator<<(std::ostream& os, const Dv::Util::Duration& u);

//! Simply compares years, months, seconds separately.
bool
operator==(const Dv::Util::Duration& u1, const Dv::Util::Duration& u2);

//! Compute Date just after period of size <code>u</code>, starting at <code>d</code>.
/*! Sometimes give strange results: e.g. 
    \code
    2000-01-31 + 1 month = 2000-03-2.
    \endcode
*/
Dv::Util::Date
operator+(const Dv::Util::Date& d, const Dv::Util::Duration& u);

#endif

