#ifndef	DV_PROPS_H
#define	DV_PROPS_H
// $Id: props.h,v 1.33 2003/10/07 12:06:45 dvermeir Exp $

/** \file
 * A \ref Dv::Util::Props class object represents a set of key-value 
 * pairs (aka property list).  Prop object can be written to and read
 * from a stream, making them suitable to implement e.g. configuration
 * files.
 *
 * The Dv::Util::env2props(Props&) function loads the shell environment 
 * (man getenv) into a Dv::Util::Props object.
*/


#include	<string>
#include	<iterator>
#include	<iostream>
#include	<map>
#include	<stdexcept> // runtime_error

namespace Dv {
namespace Util {

/**
 * Exception class associated with Props.
 * 
 * Derived from runtime_error, i.e. the only useful member function
 * is PropsException::what() which returns constant C string
 * describing the error.
 */
class PropsException: public std::runtime_error {
public:
  /**
   * Name of class, is prepended to each message argument of the
   * constructor.
   */
 static const std::string NAME;
 /**
  * Constructor, prepends NAME to message to obtain runtime_error::what().
  */
 PropsException(const std::string& message): std::runtime_error(NAME+": "+message) {}
};

/**
 * A propert list class.
 *
 * A Props object represents a mapping from key strings to value strings.
 * Values can be accessed with Props::operator[], using both
 * strings or integers as a key (the latter will first be converted
 * to strings); the resulting value will be converted to string or
 * int, as required. The result of Props::operator[]} can
 * be assigned to, thus giving a simple way to update. 
 * 
 * A Props objects can be read from and written to a stream, in
 * a format that is illustrated below (| indicates the beginning of a line):
 * 
 * \verbatim
   |database=my-database
   |query=select name, semester
   |>from course
   |host=mainserver
   |# a comment
   |age=99
   \endverbatim
 * 
 * Note that
 * <ul>
 * <li> Empty lines (with only white space chars) are ignored.
 * <li>Lines starting with '#' are ignored (useful for comments).
 * <li>A property key has the format \a X \a y where \a X is any character
 * except white space, \a ``#'' and \a ``='' and \a y is a possibly
 * empty string of characters not containing \a ``=''.
 * <li>A property key may also be empty.
 * <li>White space is copied literally, in keys as well as in values.
 * <li> A new line in a value will be represented by a \a ``>'' in the
 * first position of the following line.
 * </ul>
 * 
 */
class Props {
public:
  /**
   * A class representing a reference to a key-value pair in a Props object.
   * A non-const access with Props::operator[] returns an object
   * of this type. It supports several convertors and assignment operations
   * to support code like
   * \code
   * props["abc"] = 32;
   * \endcode
   * which will use Props::Value::operator=(int) do perform the
   * assignment.
   */
  class Value {
  public:
    /**
     * Friend class.
     */
    friend class Props;
    
    //@{
    /**
     * Props::Value::operator>>() can be used to set its reference parameter
     * with the value associated with the Props::Value object.
     * It will throw an exception, if there is no such value.
     */
    Props& operator>>(std::string&) const throw (PropsException);
    Props& operator>>(int&) const throw (PropsException);
    Props& operator>>(unsigned int&) const throw (PropsException);
    Props& operator>>(long&) const throw (PropsException);
    Props& operator>>(unsigned long&) const throw (PropsException);
    Props& operator>>(short&) const throw (PropsException);
    Props& operator>>(unsigned short&) const throw (PropsException);
    Props& operator>>(double&) const throw (PropsException);
    //@}

    /**
     * Output operator needed to avoid ambiguity
     * E.g. 
     * \code
     * cout << props["key"];
     * \endcode
     * leaves too many conversions for the compiler.
     */
    friend std::ostream& operator<<(std::ostream&,const Props::Value&);

    //@{
    /**
     * The conversion operators have a similar functionality as the
     * Props::Value::operator>>() function members.
     * Note how the string& conversion returns a reference, making
     * examples such as the following efficient.
     * \code
     * int i(props["f"]);
     * std::string& huge_string(props["text"]);
     * \endcode
     */
    operator int() const throw (PropsException);
    operator long() const throw (PropsException);
    operator short() const throw (PropsException);
    operator unsigned int() const throw (PropsException);
    operator unsigned long() const throw (PropsException);
    operator unsigned short() const throw (PropsException);
    operator double() const throw (PropsException);
    operator std::string&() throw (PropsException);
    /** Explicit "conversion" to std::string.
     * This function is handy in cases such as the following:
     * @code
     * std::string s;
     * s = props[key]; 
     * @endcode
     * which, under gcc-3.3.1 yields a compiler error due to an ambiguity between
     * @code
     * std::string::operator=(const char); 
     * std::string::operator=(const std::string); 
     * @endcode
     * and note that the conversion of Value to int provides also a conversion
     * to char (apparently of the same "quality").
     * @return reference to static_cast<std::string&>(*this)
     * @sa Dv::Util::Props::ConstValue::str
     */
    std::string& str() throw (PropsException) { return static_cast<std::string&>(*this); }
    //@}

    //@{
    /**
     * Props::Value::operator=() sets the value associated with
     * the key corresponding to Props::Value. Non-string
     * arguments are first converted to strings.
     */
    Props& operator=(int);
    Props& operator=(long);
    Props& operator=(short);
    Props& operator=(unsigned int);
    Props& operator=(unsigned long);
    Props& operator=(unsigned short);
    Props& operator=(const double&);
    Props& operator=(const std::string&);
    //@}

  private:
    Value(Props& props,const std::string& key,std::string* value): 
      props_(props), key_(key), value_(value) {}
    Value(const Value&); // forbidden cctor
    Value& operator=(const Value&); // forbidden assignment

    Props&		props_;
    const std::string	key_;
    std::string*		value_;
  };

  /**
   * A class representing a reference to a key-value pair in a Props object.
   *
   * The difference with Props::Value is that there are no assignment
   * operators (because the underlying Props object is constant.
   */
  class ConstValue {
  public:
    /**
     * Friend class.
     */
    friend class Props;

    //@{
    /**
     * Props::Value::operator>>() can be used to set its reference parameter
     * with the value associated with the Props::Value object.
     * It will throw an exception, if there is no such value.
     */
    const Props& operator>>(std::string&) const throw (PropsException);
    const Props& operator>>(int&) const throw (PropsException);
    const Props& operator>>(long&) const throw (PropsException);
    const Props& operator>>(short&) const throw (PropsException);
    const Props& operator>>(unsigned int&) const throw (PropsException);
    const Props& operator>>(unsigned long&) const throw (PropsException);
    const Props& operator>>(unsigned short&) const throw (PropsException);
    const Props& operator>>(double&) const throw (PropsException);
    //@}

    //@{
    /**
     * The conversion operators have a similar functionality as the
     * Props::Value::operator>>() function members.
     * Note how the std::string& conversion returns a reference, making
     * examples such as the following efficient.
     * \code
     * int i(props["f"]);
     * const std::string& huge_string(props["text"]);
     * \endcode
     */
    operator int() const throw (PropsException);
    operator long() const throw (PropsException);
    operator short() const throw (PropsException);
    operator unsigned int() const throw (PropsException);
    operator unsigned long() const throw (PropsException);
    operator unsigned short() const throw (PropsException);
    operator double() const throw (PropsException);
    operator const std::string&() const throw (PropsException);
    /** Explicit "conversion" to std::string.
     * This function is handy in cases such as the following:
     * @code
     * std::string s;
     * s = props[key]; 
     * @endcode
     * which, under gcc-3.3.1 yields a compiler error due to an ambiguity between
     * @code
     * std::string::operator=(const char); 
     * std::string::operator=(const std::string); 
     * @endcode
     * and note that the conversion of Value to int provides also a conversion
     * to char (apparently of the same "quality").
     * @return @a static_cast<const std::string&>(*this)
     * @sa Dv::Util::Props::Value::str
     */
    const std::string& str() const throw (PropsException) { return static_cast<const std::string&>(*this); }
    //@}
    /**
     * Output operator needed to avoid ambiguity
     * E.g. 
     * \code
     * cout << props["key"];
     * \endcode
     * leaves too many conversions for the compiler.
     */
    friend std::ostream& operator<<(std::ostream&,const Props::ConstValue&);

  private:
    ConstValue(const Props& props,const std::string& key,const std::string* value): 
      props_(props), key_(key), value_(value) {}
    ConstValue(const ConstValue&); // forbidden cctor
    ConstValue& operator=(const ConstValue&); // forbidden assignment

    const Props&	props_;
    const std::string	key_;
    const std::string*	value_;
  };

  /**
   * The underlying implementation type.
   */
  typedef  std::map<std::string,std::string*>	propmap;

  /**
   * Default constructor.
   */
  Props() {}
  /**
   * Reads value=key pairs from istream.
   *
   * \param is istream from which \a key=value pairs are are read.
   *
   * This constructor uses Props::read(\a istream&, \a -1) to read
   * the rest of the open istream into a Props object.
   * This constructor is convenient to read in configuration files
   * in a format as described above.
   * \code
   * ifstream ifs("/usr/local/share/z.config");
   * Props config(ifs);
   * try {
   *   int max = config["max"];
   *   }
   * catch (PropsException& e) {
   *   cerr << "Configuration file does not define 'max'" << endl;
   *   exit(1);
   *   }
   * \endcode
   * \warning
   * Note that this constructor does not work properly with streams
   * that have no clear ``\a eof'' indication, e.g. simple network connections.
   * \see Props
   */
  Props(std::istream& is); 
  /**
   * Virtual destructor.
   */
  virtual ~Props();
  //@{
  /**
   * Key-based ``r-value'' access to properties.
   * \param key value, can be of a variety of types, it will be
   *   converted to std::string.
   * \return a Props::ConstValue which can be converted to 
   *   a variety of types.
   * \exception PropsException if no value for \a key can be found.
   * \see Props::ConstValue::operator int
   *
   * This overloaded member functions supports, together with the
   * conversion functions for Props::ConstValue, simple retrieval
   * a value corresponding to a given key, as illustrated in the
   * following example.
   * \code
   * props["name"] = "lisa";
   * props[0] = "zero";
   * props[10] = 11;
   * props["age"] = 21;
   *  
   * int age(props["age"]);
   * std::string name(props["name"]);
   * int n(props[10]);
   * \endcode
   */
  ConstValue operator[](const std::string& key) const throw (PropsException);
  ConstValue operator[](int key) const throw (PropsException);
  ConstValue operator[](unsigned int key) const throw (PropsException);
  ConstValue operator[](long key) const throw (PropsException);
  ConstValue operator[](unsigned long key) const throw (PropsException);
  ConstValue operator[](short key) const throw (PropsException);
  ConstValue operator[](unsigned short key) const throw (PropsException);
  ConstValue operator[](const double& key) const throw (PropsException);
  //@}
  //@{
  /**
   * Key-based l-value access to properties.
   * \param k key value, can be of a variety of types, it will be
   *   converted to std::string.
   * \return a Props::Value object which can be assigned to.
   *
   * These member functions support, together with the conversion and assignment
   * functions for Props::Value simple retrieval and assignment of a value
   * corresponding to a given \a key, as illustrated in the following example.
   *
   * \code
   * props["name"] = "lisa";
   * props[0] = "zero";
   * props[10] = 11;
   * props["age"] = 21;
   *  
   * int age(props["age"]);
   * std::string name(props["name"]);
   * int n(props[10]);
   * \endcode
   */
  Value	operator[](const std::string& k);
  Value	operator[](int k);
  Value	operator[](unsigned int k);
  Value	operator[](long k);
  Value	operator[](unsigned long k);
  Value	operator[](short k);
  Value	operator[](unsigned short k);
  Value	operator[](const double& k);
  //@}

  /**
   * Erase any value that is associated with key.
   * \param key for which the associated \a key=\a value pair will be
   *   removed.
   */
  Props& erase(const std::string& key);
  /**
   * Erase all \a key = \a value pairs from the property list.
   * \see Props::erase
   */
  Props& clear() { props_.clear(); return *this; }

  /**
   * Write the Props object to the stream, in the format described above.
   */
  friend std::ostream& operator<<(std::ostream&,const Props &);

  /**
   * Add a \a key = \a value pair.
   * Use Props::operator[] for automatic type conversion to std::string.
   *
   * \param key std::string.
   * \param value std::string.
   * \return reference to \a *this.
   * \warning
   * The value added overrides any existing \a value for \a key.
   * \see Props::operator[]
   *
   */
  Props& add(const std::string& key,const std::string& value);
  /**
   * Append to a value associated with a \a key.
   * The \a value, preceded by the \a separator, is appended to the current
   * value associated with \a key.
   * If there is no current value, Props::append behaves like Props::add.
   *
   * \param key for which the value is to be extended
   * \param value to append to the current value of \a key
   * \param separator is appended to \a (*this)[key] before \a value,
   *  if \a (*this)[key] was not yet defined.
   * \see Props::add
   */
  Props& append(const std::string& key, const std::string& value, 
  	        const std::string& separator=" ");

  /**
   * Try to find value for key, return null if not found.
   * \param key to search for.
   * \return pointer to \a value associated with \a key or 0 if none
   *   found.
   */
  const std::string* find(const std::string& key) const;
  /**
   * Try to find value for key, return null if not found.
   * \param key to search for.
   * \see Props::find
   */
  const std::string* operator()(const std::string& key) const { return find(key); }

  /**
   * Map iterator, *iterator is a pair<std::string,*std::string>.
   */
  typedef  propmap::const_iterator iterator;
  /**
   * \return iterator referring to the first \a key=value pair.
   */
  iterator begin() const { return props_.begin(); }
  /**
   * \return iterator referring past the last  \a key=value pair.
   */
  iterator end() const { return props_.end(); }
  /**
   * \return number of \a key=value pairs in the property list.
   */
  unsigned size() const	{ return props_.size(); }

  /** 
   * Read \a key-value pairs from istream until the end of the stream
   *   is reached or a syntax error occurs.
   * \param is stream to read from.
   * \param props property object to update.
   * \return reference to \a is.
   * \warning
   *  Existing key values will be overriden.
   * \see Props::read
   */
  friend std::istream& operator>>(std::istream& is, Props& props);

  /**
   * Read \a n \a key=value pairs from istream, or until \a eof 
   * if \a n <\a 0.
   * Note that the Props object is not cleared before reading
   * the \a key=value pairs, they are simply added or override
   * existing values.
   * \param is stream to read from.
   * \param n number of \a key=value pairs to read. If \a n<0, reading
   *          will continue until \a eof.
   * \return reference to is.
   */
  std::istream&	read(std::istream& is, int n=-1);

  /**
   * Return a reference to the \a key=value pair with the given key.
   * This is an auxiliary function. It supports expressions such as
   * the following.
   *
   *  \code
   *  int   value;
   *  Props props;
   *  props >> "key" >> value;
   *  \endcode
   *
   * \param key to search for.
   * \return a ConstValue which can be converted to any type using
   *  ConstValue::operator>>.
   * \warning The conversion from the Props::ConstValue will throw
   *   an exception if there was no \a value corresponding to \a key.
   * \see ConstValue::operator>>
   */
  ConstValue operator>>(const char* key) const { 
	return ConstValue(*this,key,find(key)); }

  /**
   * Recursively substitute variable occurrences in sin to yield sout.
   * \param sin stream that is appended to \a sout.
   * \param sout stream that is appended to.
   * \param f function that transforms strings. ${name}
   *	in \a sin is replaced by \a f((*this)[name]).
   * \exception invalid_argument is thrown upon a syntax error.
   *
   * Sin is a stream, the contents which is appended to sout except for variable
   * occurrences of the form ${key} or %{key} which 
   * are replaced by \a f((*this)[key]) and \a (*this)[key],
   * respectively (the ``$'' sign can be escaped using ``\'').
   *  
   * Undefined (in *this) variables are replaced by the empty std::string.
   *
   * Example:
   * \code
   * std::string query("select * from");
   * props.substitute("person where name = '${name}' and age = %{age}",
   *		query,MySql::escape);
   * \endcode
   *
   * Syntax:
   * <ul>
   * <li> Variables start with ``${'' or ``%{'' except if the
   *      ``$'' or ``%'' is preceeded by a backslash (``\'').
   * <li> A backslash is a normal character, except if it
   *      is immediately followed by a ``$'' or a ``%'', in which
   *      case it is not copied.
   * </ul>
   */
  void substitute(std::istream& sin,std::ostream& sout,
        std::string (*f)(const std::string&) = 0) const throw (std::invalid_argument);
  /**
   * Recursively substitute variable occurrences in sin to yield sout.
   *
   * \param sin std::string that is appended to \a sout.
   * \param sout std::string that is appended to.
   * \param f function that transforms strings. ${name}
   *	in \a sin is replaced by \a f((*this)[name]).
   * \exception invalid_argument is thrown upon a syntax error.
   *
   * Sin is a std::string which is appended to sout except for variable
   * occurrences of the form ${key} or %{key} which 
   * are replaced by \a f((*this)[key]) and \a (*this)[key],
   * respectively (the ``$'' sign can be escaped using ``\'').
   *  
   * Undefined (in \a *this) variables are replaced by the empty std::string.
   *
   * Syntax:
   * <ul>
   * <li> Variables start with ``${'' or ``%{'' except if the
   *      ``$'' or ``%'' is preceeded by a backslash (``\'').
   * <li> A backslash is a normal character, except if it
   *      is immediately followed by a ``$'' or a ``%'', in which
   *      case it is not copied.
   * </ul>
   */
  void substitute(const std::string& sin, std::string& sout,
        std::string (*f)(const std::string&) = 0) const throw (std::invalid_argument);
  /**
   * Same as above but returns result std::string.
   * \param sin std::string that is appended to return value.
   * \param f function that transforms strings. ${name}
   *	in \a sin is replaced by \a f((*this)[name]).
   * \return std::string with subsitutions performed.
   * \exception invalid_argument is thrown upon a syntax error.
   * \see Props::substitute
   */
  std::string substitute(const std::string& sin, std::string (*f)(const std::string&) = 0)
    const throw (std::invalid_argument);

  /**
   * Insert selected \a subkey=value pairs of this object into the target.
   * The selected pairs are those for which the \a key starts with
   * the \a scope std::string followed by '\a ::'. The pairs will be inserted
   * with a \a key consisting of the remainder of the original key,
   * after removing the \a scope std::string and '::'.
   *
   * Example.
   * \code
   * Props	config;
   * Props	dbconfig;
   * config.select("database",dbconfig);
   * \endcode
   *  
   * In the example, if <tt>config["database::name"] == "admindb"</tt>
   * then <tt>dbconfig["name"] == "admindb"</tt> after the call to
   * <tt>select</tt>.
   *
   * The function returns a reference to its second first argument.
   * 
   * \param target the Props object the selected pairs will be added to.
   * \param scope the selected pairs have ``<tt>scope::</tt>'' as a key. 
   * \return a reference to \a target.
   */
  Props& select(const std::string& scope,Props& target) const;

  /**
   * Load all \a key-value pairs from another props object under a certain
   * scope.
   * All key-value pairs from \a source will be added to \a *this,
   * where the key will be prefixed with <tt>scope::</tt> if the size
   * of \a scope is not zero.
   *
   * Example usage:
   * \code
   * Props	dbconfig;
   * Props	config;
   * config.merge(dbconfig,"database");
   * \endcode
   *
   * If <tt>dbconfig["name"]=="admindb"</tt> then
   * <tt>config["database::name"]=="admindb"</tt> after the merge.
   *
   * \param source the Props object from which \a key=value pairs will
   *   be copied.
   * \param scope only \a key=value pairs where \a key starts with \a scope::
   *   will be copied.
   * \return reference to \a *this.
   * \warning \a &source must be different from \a this.
   * \see Dv::Util::Props::select
   */
  Props& merge(const Props& source,const std::string& scope="") throw (PropsException);
  /**
   * Compare two Props objects.
   */
  bool operator==(const Props&) const;
  bool operator!=(const Props& p) const { return !(*this == p); }

  /**
   * Send a props object over a stream.
   * The difference with e.g. 
   * \code
   * ostream& operator<<(ostream&,const Props::Value&);
   * \endcode
   * is that Props::send also works over network connections where
   * the end of the input is not indicated by \a EOF.
   * An object sent by Props::send() should be received
   * by Props::receive().
   * \param os stream to write Props object to.
   * \see Dv::Util::Props::receive
   */
   std::ostream& send(std::ostream& os) const;
   /**
    * Receive a props object from a stream.
    * The difference with e.g. 
    * \code
    * Props(istream&); 
    * \endcode
    * is that Props::receive also works over network connections where
    * the end of the input is not indicated by \a EOF.
    * An object received by Props::receive() should have been sent
    * by Props::send().
    *
    * \param is stream from which to read the Props object.
    * \warning
    * Note that the object is not cleared before receiving new
    * \a key=value pairs.
    * \see Dv::Util::Props::send
    */
   std::istream& receive(std::istream& is);
  /**
   * Copy construction is allowed.
   */
  Props(const Props& props) throw (PropsException);
  /**
   * Assignment allowed.
   */
  Props& operator=(const Props& props) throw (PropsException);
private:
  // Try to find value for key, return null if not found
  std::string* found(const std::string& key);

  // Using a value type of std::string* allows us to return a reference
  // to a value (operator[]) [values don't move], which is good
  // for efficiency when retrieving large values.
  propmap  	props_;
};

/**
 * This function adds its environment to its props parameter.
 * It returns a reference to its parameter.
 * See man (5) environ for information about the environment.
 * \param props to add environment variables to.
 * \return reference to \a props.
 */
Props& env2props(Props& props);

}}
#endif
