#ifndef DVUTIL_PSTREAM_H
#define DVUTIL_PSTREAM_H
// $Id: pstream.h,v 1.5 2003/09/19 18:08:43 dvermeir Exp $

#include <string>
#include <iostream>
#include <stdexcept>

namespace Dv {
namespace Util {

/** A wrapper that makes a stream of popen(command,"r").  */
class pstream: public std::iostream {
public:
  /** Constructor.
   * Create an input or output stream for data produced by a command. If mode is
   * "r" (read), reading the stream reads output from the command.
   * If mode is "w" (write), writing the stream sends the data
   * to the command's standard input.
   * @param command see popen(3)
   * @param mode read ("r") or write ("w"), see popen(3).
   * @exception std::runtime_error if anything goes wrong.
   * @warning E.g. for a stream opened in read mode, the default
   *   standard ouput of the command is still the standard output of the
   *   parent process. 
   */
  pstream(const std::string& command, const std::string& mode) throw (std::runtime_error);

  /** Close stream.
   * Wait for command to finish, then return its exit status.
   * After a close, the stream is unusable.
   * @return exit status of command as returned by main().
   * @warning A negative exit status gets converted to a
   *   positive number, i.e. -1 becomes 255.
   * @exception std::runtime_error if command did not exit normally or any other problem.
   */
  size_t close() throw (std::runtime_error);

  /** Destructor.
   * @exception std::runtime_error if anything goes wrong.
   */
  ~pstream() throw (std::runtime_error);

  /* @return command producing data for this stream */
  const std::string& command() const { return command_; }
  /* @return mode of this stream */
  const std::string& mode() const { return mode_; }
private:
  void* file_; // opaque stdio FILE*
  const std::string command_;
  const std::string mode_;
};
}}
#endif
