#ifndef	DV_REF_H
#define	DV_REF_H
// $Id: ref.h,v 1.15 2003/07/14 09:15:28 dvermeir Exp $

/*! \file 
 * The Dv::ref class implements reference-counted pointers.  
 */

namespace Dv {
namespace Util {
//!  Reference-counted pointers.
/*!  Example usage:

  \code
  #include	<iostream>
  #include	<string>
  #include	<dvutil/ref.h>
  using Dv::ref;	

  class Huge {
  friend class ref<Huge>;
  public:
    friend ostream& operator<<(ostream& os, const Huge& h) { 
      return os << h._s; }
    const string& s() const { return _s; }
    static Dv::ref<Huge> create(const char *s) { 
      return ref<Huge>(new Huge(s)); }
  private:
    string		_s;
    Huge(const char* s): _s(s) {}
    ~Huge();
  };
  
  int
  main(int argc,char *argv[]) {
  ref<Huge>	r = Huge::create("c++");
  ref<Huge>	p = r;
  cout << *p << p->s() << endl;
  }
  \endcode
*/
template<class T>
class ref {
//! only T can produce non-null ref<T> objects.
// friend T;
// template<typename X> friend class ref; // does not work with gcc 2.95.3
public:
  /** Default constructor makes null ref<T> */
  ref(): t_(0), n_(new unsigned int(1)) { /* cerr << "0-CTOR" << endl; */}
  //! copy constructor increments reference count
  ref(const ref<T>& r): t_(r.operator->()), n_(r.n()) { ++*n_; }

  template<typename X>
  explicit ref(const ref<X>& r): t_(r.operator->()), n_(r.n()) { ++*n_; }
  //! destructor decrements reference count or deletes underlying pointer
  ~ref() { if (--*n_) return; delete t_; delete n_; }

  //! return pointer to counter
  // should be private but gcc 2.95.3 does not accept template<typename X> friend class ref;
  unsigned int*  n() const  { return n_; }
  //! is this a null ref<T>?
  operator  bool() const  { return (t_!=0); }
  //! ref<T> objects can be dereferenced like a pointer
  T&  operator*() const { return *t_; }
  //! ref<T> objects support -> like pointers
  T*  operator->() const { return t_; }    
  //! assignment decrements reference count of lhs, increases count for rhs
  ref<T>&  operator=(const ref<T>& r) { 
             if (this==&r) return *this;
             if (--*n_==0) { delete t_; delete n_; }
             t_ = r.t_;
	     n_=r.n_;
	     ++*n_; 
	     return *this;  
	     }
  //! == compares underlying T objects
  bool operator==(const ref<T>& r) const {
        if (t_ == r.t_) return true;
        if (t_)
          if (r.t_)
            return (*t_ == *r);
          else
            return false;
        // r.t_ != t_ == 0
        return false;
        }
  //! < compares underlying T objects
  bool operator<(const ref<T>& r) const {
        // null is minimal element in this implementation
        if (t_ == r.t_) return false;
        if (t_)
          if (r.t_)
            return (*t_ < *r);
          else
            return false;
        // t_ == 0, r.t_ != 0
        return true;
        }
// protected:
  //! only T objects can encapsulate a (new'ed) pointer into a ref<T>
  // template<typename X>
  explicit ref(T* t): t_(t), n_(new unsigned int(1)) {}

  template<typename X>
  operator ref<X>() const { return ref<X>(*this); }
  
private:
  
  T*    	t_;
  unsigned int*	n_;
};
}}

#endif
