#ifndef	DV_USER_H
#define DV_USER_H
// $Id: user.h,v 1.7 2002/10/18 13:17:18 dvermeir Exp $

#include	<sys/types.h> // uid_t
#include	<stdexcept> // runtime_error

/*! \file
  A \ref Dv::Util::User class object represents information about a Unix user, as obtained
  by the getpwnam() function.
*/

namespace Dv {
namespace Util {

//! Exception class, thrown by some User function members.
class UserError: public std::runtime_error {
public:
  //! Name of this class.
  static const std::string NAME;
  //! Prepend NAME to message to obtain runtime_error::what().
  UserError(const std::string& message): std::runtime_error(NAME+": "+message) {}
};

//! A class representing a Unix user.
/*! This class provides a convenient wrapper around information that
    is usually obtained using getpwnam().
*/
class User {
public:
  //! Initialize to current user, equivalent with User(getuid()).
  User();
  //! Initialize to user with given uid, throw UserError if not found.
  User(uid_t uid) throw (UserError);
  //! Initialize to user with given user name, throw UserError if not found.
  User(const std::string& name) throw (UserError);
  
  //! Retrieve uid of this user.
  uid_t		uid() const { return uid_; }
  //! Retrieve current group id of this user.
  gid_t		gid() const { return gid_; }
  //! Retrieve pathname of home directory of this user.
  const std::string&	home() const { return home_; }
  //! Retrieve pathname of shell program of this user.
  const std::string&	shell() const { return shell_; }
  //! Retrieve user name of this user.
  const std::string&	name() const { return name_; }
  //! Retrieve full name (gecos field) of this user.
  const std::string&	full_name() const { return full_name_; }
  //! Return true iff called by root and password matches user's password.
  bool password(const std::string& pass) const;

  //! Users are equal if they have the same uid.
  bool operator==(const User& u) const { return uid() == u.uid(); }
  //! Based on the comparison of the uid's.
  bool operator<(const User& u) const { return uid() < u.uid(); }

  bool operator>(const User& u) const	{ return u < *this; }
  bool operator!=(const User& u) const	{ return ! (u == *this); }
  bool operator>=(const User& u) const	{ return ! (*this < u); }
  bool operator<=(const User& u) const	{ return ! (*this > u); }

private:
  void*		getpwinfo(void*);

  std::string	name_;
  std::string	full_name_;
  uid_t		uid_;
  gid_t		gid_;
  std::string	home_;
  std::string	shell_;
};

}}
#endif
